using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Extended;

namespace H3UtilityToolkit.Essentials
{
    /// <summary>
    /// Object for reading, storing, and editing data from the String Table.
    /// </summary>
    public class Strings
    {
        #region Instance Data

        protected Map __map__;
        private string[] __stringsList__;
        private int[] __offsetsList__;

        #endregion

        #region Accessors

        ///<summary>
        /// Map
        ///</summary>
        protected Map Map { get { return __map__; } set { __map__ = value; } }
        ///<summary>
        /// StringsList
        ///</summary>
        public string[] StringsList { get { return __stringsList__; } set { __stringsList__ = value; } }
        ///<summary>
        /// OffsetsList
        ///</summary>
        public int[] OffsetsList { get { return __offsetsList__; } set { __offsetsList__ = value; } }

        #endregion

        /// <summary>
        /// Initialises a new instance of the Strings class based on the
        /// supplied Map reference.
        /// </summary>
        /// <param name="map">The supplied Map reference. </param>
        public Strings(Map map)
        {
            this.Map = map;
        }

        /// <summary>
        /// Initialises a new instance of the Strings class based on the
        /// supplied Map reference, and begins reading straight away with
        /// the supplied BitReader.
        /// </summary>
        /// <param name="map">The supplied Map reference. </param>
        /// <param name="br">The supplied BitReader. </param>
        public Strings(Map map, BitReader br)
        {
            this.Map = map;
            this.Read(br);
        }

        /// <summary>
        /// Reads data from the String Table and stores it for access later.
        /// </summary>
        public void Read()
        {
            BitReader br = new BitReader(Map.FileName, Endian.Big);
            this.Read(br);
            br.Close();
        }

        /// <summary>
        /// Reads data from the String Table and stores it for access later
        /// with the supplied BitReader.
        /// </summary>
        /// <param name="br">The supplied BitReader. </param>
        public void Read(BitReader br)
        {
            // Gather the offsets
            if (Map.Header.StringTableIndexOffset >= Header.Length && Map.Header.StringTableIndexOffset < br.BaseStream.Length)
            {
                OffsetsList = new int[Map.Header.StringTableCount];
                br.BaseStream.Position = Map.Header.StringTableIndexOffset;
                for (int x = 0; x < Map.Header.StringTableCount; x++)
                {
                    OffsetsList[x] = br.ReadInt32();
                }
            }
            else OffsetsList = new int[0];

            // Gather the strings
            if (Map.Header.StringTableOffset >= Header.Length && Map.Header.StringTableOffset < br.BaseStream.Length)
            {
                StringsList = new string[Map.Header.StringTableCount];
                int indx;
                for (indx = 0; indx < Map.Header.StringTableCount - 1; indx++)
                {
                    br.BaseStream.Position = Map.Header.StringTableOffset + OffsetsList[indx];
                    StringsList[indx] = br.ReadString(OffsetsList[indx + 1] - OffsetsList[indx] - 1);
                }
                br.BaseStream.Position = Map.Header.StringTableOffset + OffsetsList[indx];
                StringsList[indx] = br.ReadString(Map.Header.StringTableSize - OffsetsList[indx] - 1);
            }
            else StringsList = new string[0];
        }

        /// <summary>
        /// Parses the data read from the String Table into a ListViewItem array.
        /// </summary>
        public ListViewItem[] ParseForListView()
        {
            ListViewItem[] items = new ListViewItem[StringsList.Length];
            for (int x = 0; x < StringsList.Length; x++)
            {
                items[x] = new ListViewItem(StringsList[x]);
                items[x].SubItems.Add(OffsetsList[x].ToString());
                items[x].SubItems.Add(StringsList[x].Length.ToString());
            }
            return items;
        }

        /// <summary>
        /// Writes the string at the specified index to the String Table.
        /// </summary>
        /// <param name="index">The specified index. </param>
        public void Write(int index)
        {
            BitWriter bw = new BitWriter(Map.FileName, Endian.Big);
            this.Write(bw, index);
            bw.Close();
        }

        /// <summary>
        /// Writes the string at the specified index to the String Table
        /// using the supplied BitWriter.
        /// </summary>
        /// <param name="bw">The supplied BitWriter. </param>
        /// <param name="index">The specified index. </param>
        public void Write(BitWriter bw, int index)
        {
            bw.BaseStream.Position = OffsetsList[index];
            bw.Write(StringsList[index]); // (should be same size as original)
        }

        /// <summary>
        /// Writes all strings stored in the class to the String Table.
        /// </summary>
        public void WriteAll()
        {
            BitWriter bw = new BitWriter(Map.FileName, Endian.Big);
            this.WriteAll(bw);
            bw.Close();
        }

        /// <summary>
        /// Writes all strings stored in the class to the String Table
        /// using the supplied BitWriter.
        /// </summary>
        /// <param name="bw">The supplied BitWriter. </param>
        public void WriteAll(BitWriter bw)
        {
            for (int x = 0; x < Map.Header.StringTableCount; x++)
            {
                this.Write(bw, x);
            }
        }
    }
}