using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Extended;

namespace H3UtilityToolkit.Essentials
{
    /// <summary>
    /// Object for reading, storing, and editing data relative to the Tag structures
    /// </summary>
    public class Tags
    {
        #region Instance Data

        #region Constants

        /// <summary>
        /// [Const] NullTagType
        /// </summary>
        public const string NullTagType = "";
        /// <summary>
        /// [Const] NullTagPath
        /// </summary>
        public const string NullTagPath = "Nulled Out";

        #endregion

        protected Map __map__;
        private string[] __tagClasses__;
        private string[] __parentClasses__;
        private string[] __grandParentClasses__;
        private Tag[] __tagsList__;
        private SortTagsBy __sortTagsBy__;
        private int __selectedIndex__;

        #endregion

        #region Accessors

        ///<summary>
        /// Map
        ///</summary>
        protected Map Map { get { return __map__; } set { __map__ = value; } }
        ///<summary>
        /// TagClasses
        ///</summary>
        public string[] TagClasses { get { return __tagClasses__; } protected set { __tagClasses__ = value; } }
        ///<summary>
        /// ParentClasses
        ///</summary>
        public string[] ParentClasses { get { return __parentClasses__; } protected set { __parentClasses__ = value; } }
        ///<summary>
        /// GrandParentClasses
        ///</summary>
        public string[] GrandParentClasses { get { return __grandParentClasses__; } protected set { __grandParentClasses__ = value; } }
        ///<summary>
        /// TagsList
        ///</summary>
        public Tag[] TagsList { get { return __tagsList__; } protected set { __tagsList__ = value; } }
        ///<summary>
        /// SortTagsBy
        ///</summary>
        public SortTagsBy SortTagsBy { get { return __sortTagsBy__; } set { __sortTagsBy__ = value; } }
        ///<summary>
        /// SelectedIndex
        ///</summary>
        public int SelectedIndex { get { return __selectedIndex__; } set { __selectedIndex__ = value; } }
        ///<summary>
        /// SelectedTag
        ///</summary>
        public Tag SelectedTag { get { return TagsList[SelectedIndex]; } }

        #endregion

        /// <summary>
        /// Initialises a new instance of the Tags class based on the
        /// supplied Map reference.
        /// </summary>
        /// <param name="map">The supplied Map reference. </param>
        public Tags(Map map)
        {
            this.Map = map;
            this.SelectedIndex = -1;
            this.SortTagsBy = SortTagsBy.TagClass;
        }

        /// <summary>
        /// Initialises a new instance of the Tags class based on the
        /// supplied Map reference, and begins reading straight away with
        /// the supplied BitReader.
        /// </summary>
        /// <param name="map">The supplied Map reference. </param>
        /// <param name="br">The supplied BitReader. </param>
        public Tags(Map map, BitReader br)
        {
            this.Map = map;
            this.SelectedIndex = -1;
            this.SortTagsBy = SortTagsBy.TagClass;
            this.Read(br);
        }

        /// <summary>
        /// Reads data relevant to the Tag structures and stores it for access later.
        /// </summary>
        public void Read()
        {
            BitReader br = new BitReader(Map.FileName, Endian.Big);
            this.Read(br);
            br.Close();
        }

        /// <summary>
        /// Reads data relevant to the Tag structures and stores it for access later
        /// with the supplied BitReader.
        /// </summary>
        /// <param name="br">The supplied BitReader. </param>
        public void Read(BitReader br)
        {
            // Index - TagTypes
            if (Map.Index.TagClassIndexOffset >= Header.Length && Map.Index.TagClassIndexOffset < br.BaseStream.Length)
            {
                TagClasses = new string[Map.Index.TagClassCount];
                ParentClasses = new string[Map.Index.TagClassCount];
                GrandParentClasses = new string[Map.Index.TagClassCount];

                br.BaseStream.Position = Map.Index.TagClassIndexOffset;
                for (int x = 0; x < Map.Index.TagClassCount; x++)
                {
                    TagClasses[x] = br.ReadString(4);
                    ParentClasses[x] = br.ReadString(4);
                    GrandParentClasses[x] = br.ReadString(4);
                    br.BaseStream.Position += 4;
                }
            }
            else
            {
                TagClasses = new string[0];
                ParentClasses = new string[0];
                GrandParentClasses = new string[0];
            }

            // Index - Tag Info
            if (Map.Index.TagInfoOffset >= Header.Length && Map.Index.TagInfoOffset < br.BaseStream.Length)
            {
                TagsList = new Tag[Map.Index.TagCount];
                br.BaseStream.Position = Map.Index.TagInfoOffset;
                for (int x = 0; x < Map.Index.TagCount; x++)
                {
                    TagsList[x] = new Tag();

                    short classIndex = br.ReadInt16();
                    if (classIndex < 0 || classIndex >= TagClasses.Length) TagsList[x].Type = NullTagType;
                    else TagsList[x].Type = TagClasses[classIndex];

                    TagsList[x].Identity = br.ReadInt16();
                    TagsList[x].Identity <<= 16;
                    TagsList[x].Identity |= x;
                    TagsList[x].RawMetaOffset = br.ReadInt32();
                    TagsList[x].MetaOffset = TagsList[x].RawMetaOffset - Map.Header.MapMagic;
                    TagsList[x].MetaSize = GetMetaSize(TagsList[x].Type);
                }
            }
            else TagsList = new Tag[0];

            // Paths
            if (Map.Header.FileTableOffset < Header.Length || // Map is protected ;)
                Map.Header.FileTableOffset > br.BaseStream.Length)
            {
                for (int x = 0; x < Map.Index.TagCount; x++)
                {
                    TagsList[x].PathOffset = 0;
                    TagsList[x].PathLength = NullTagPath.Length;
                    TagsList[x].Path = NullTagPath;
                }
            }
            else
            {
                // File Index
                br.BaseStream.Position = Map.Header.FileTableIndexOffset;
                int[] offsets = new int[Map.Index.TagCount + 1];
                int indx;
                for (indx = 0; indx < Map.Index.TagCount; indx++)
                {
                    offsets[indx] = br.ReadInt32();
                }
                offsets[indx] = Map.Header.FileTableSize;

                // File Table
                br.BaseStream.Position = Map.Header.FileTableOffset;
                for (int x = 0; x < Map.Index.TagCount; x++)
                {
                    TagsList[x].PathOffset = offsets[x];
                    TagsList[x].PathLength = offsets[x + 1] - offsets[x] - 1;
                    br.BaseStream.Position = Map.Header.FileTableOffset + offsets[x];
                    TagsList[x].Path = br.ReadString(TagsList[x].PathLength);
                }
            }

            /*
            // Meta Sizes - will be in-accurate, but anyways..

            int endOfMeta = Map.Header.Filesize;
            int matgIndx = IndexOf("matg", "globals\\globals");
            if (matgIndx != -1)
            {
                br.BaseStream.Position = TagsList[matgIndx].MetaOffset + 452 + 8;
                endOfMeta = br.ReadInt32() + Map.Header.LocaleTableAddrMod;
            }

            for (int x = 0; x < Map.Index.TagCount; x++)
            {
                int curOffset = endOfMeta;
                for (int y = 0; y < Map.Index.TagCount; y++)
                {
                    if (x != y &&
                        TagsList[y].MetaOffset > TagsList[x].MetaOffset &&
                        TagsList[y].MetaOffset < curOffset)
                    {
                        curOffset = TagsList[y].MetaOffset;
                    }
                }

                // The Meta is split into two groups, this way we won't get the entire middle bit
                // in the last meta of the last tag in the first group.
                if (TagsList[x].MetaOffset < Map.Index.TagInfoHeaderOffset && curOffset > Map.Index.TagInfoHeaderOffset)
                {
                    curOffset = Map.Index.TagInfoHeaderOffset;
                }

                // Finally get size.
                TagsList[x].MetaSize = curOffset - TagsList[x].MetaOffset;
            }
            */
        }

        /// <summary>
        /// Gets the size of the meta for the specified type.
        /// </summary>
        /// <param name="type">The specified type. </param>
        public static int GetMetaSize(string type)
        {
            switch (type)
            {
                case "_fx_": return 28;
                case "adlg": return 76;
                case "ant!": return 76;
                case "beam": return 12;
                case "bipd": return 1248;
                case "bitm": return 164;
                case "bkey": return 84;
                case "bloc": return 268;
                case "bmp3": return 92;
                case "bsdt": return 160;
                case "cddf": return 40;
                case "cfxs": return 228;
                case "chad": return 92;
                case "chdt": return 24;
                case "chgd": return 240;
                case "chmt": return 16668;
                case "clwd": return 148;
                case "cntl": return 12;
                case "coll": return 68;
                case "colo": return 12;
                case "crea": return 504;
                case "ctrl": return 460;
                case "dctr": return 128;
                case "decs": return 36;
                case "draw": return 4;
                case "drdf": return 12;
                case "dsrc": return 28;
                case "effe": return 104;
                case "effg": return 12;
                case "eqip": return 640;
                case "flck": return 92;
                case "foot": return 12;
                case "glps": return 28;
                case "glvs": return 28;
                case "grup": return 92;
                case "hlmt": return 392;
                case "jmad": return 272;
                case "jmrq": return 12;
                case "jpt!": return 328;
                case "lens": return 152;
                case "ligh": return 148;
                case "lsnd": return 64;
                case "lst3": return 112;
                case "ltvl": return 12;
                case "mach": return 424;
                case "matg": return 58232;
                case "mffn": return 56;
                case "mode": return 460;
                case "mulg": return 24;
                case "perf": return 12;
                case "phmo": return 408;
                case "pixl": return 32;
                case "play": return 28584;
                case "pmdf": return 144;
                case "pmov": return 32;
                case "pphy": return 148;
                case "proj": return 672;
                case "prt3": return 404;
                case "rasg": return 164;
                case "rmd ": return 64;
                case "rmdf": return 92;
                case "rmfl": return 68;
                case "rmhg": return 68;
                case "rmop": return 12;
                case "rmsh": return 68;
                case "rmt2": return 132;
                case "rmtr": return 92;
                case "rmw ": return 64;
                case "rwrd": return 168;
                case "sbsp": return 904;
                case "scen": return 256;
                case "scn3": return 168;
                case "scnr": return 2276;
                case "sddt": return 64;
                case "sefc": return 12;
                case "sfx+": return 12;
                case "sgp!": return 72;
                case "shit": return 164;
                case "skn3": return 28;
                case "skya": return 76;
                case "sLdT": return 88;
                case "sncl": return 12;
                case "snd!": return 32;
                case "snde": return 72;
                case "snmx": return 220;
                case "spk!": return 40;
                case "ssce": return 264;
                case "styl": return 92;
                case "trak": return 16;
                case "txt3": return 60;
                case "udlg": return 36;
                case "ugh!": return 184;
                case "uise": return 320;
                case "unic": return 80;
                case "vehi": return 1100;
                case "vtsh": return 32;
                case "wacd": return 576;
                case "wclr": return 36;
                case "weap": return 1280;
                case "wezr": return 136;
                case "wgan": return 120;
                case "wgtz": return 60;
                case "wigl": return 352;
                case "wind": return 120;
                case "wpos": return 36;
                case "wrot": return 36;
                case "wscl": return 36;
                case "wspr": return 36;
                case "zone": return 532;
                default: return 0;
            }
        }

        /// <summary>
        /// Finds Tag index based on the supplied identifier.
        /// </summary>
        /// <param name="ID">The supplied identifier. </param>
        public int IndexOf(int ID)
        {
            for (int x = 0; x < TagsList.Length; x++)
            {
                if (TagsList[x].Identity == ID)
                    return x;
            }
            return -1;
        }

        /// <summary>
        /// Finds Tag index based on the supplied type and path.
        /// </summary>
        /// <param name="type">The supplied type. </param>
        /// <param name="path">The supplied path. </param>
        public int IndexOf(string type, string path)
        {
            for (int x = 0; x < TagsList.Length; x++)
            {
                if (TagsList[x].Type == type && TagsList[x].Path == path)
                    return x;
            }
            return -1;
        }

        /// <summary>
        /// Finds Tag index based on the supplied type. First match is returned.
        /// </summary>
        /// <param name="type">The supplied type. </param>
        public int FirstIndexOf(string type)
        {
            for (int x = 0; x < TagsList.Length; x++)
            {
                if (TagsList[x].Type == type)
                    return x;
            }
            return -1;
        }

        /// <summary>
        /// Parses the tag data of the specified type into a TreeNode array, 
        /// also removing empty classes if specified.
        /// </summary>
        /// <param name="type">The specified type. </param>
        /// <param name="removeEmptyClasses">True to remove classes which contain no tags. </param>
        public TreeNode[] ParseForTreeview(string type, bool removeEmptyClasses)
        {
            // Load all
            if (type == string.Empty)
            {
                switch (SortTagsBy)
                {
                    case SortTagsBy.TagClass:
                        {
                            if (removeEmptyClasses)
                            {
                                List<TreeNode> nodes = new List<TreeNode>(TagClasses.Length);
                                for (int x = 0; x < TagClasses.Length; x++)
                                {
                                    TreeNode node = new TreeNode(TagClasses[x], 1, 1);
                                    for (int y = 0; y < TagsList.Length; y++)
                                    {
                                        if (TagsList[y].Type == TagClasses[x])
                                            node.Nodes.Add(string.Empty, TagsList[y].Path, 2, 2);
                                    }
                                    if (node.Nodes.Count != 0) nodes.Add(node);
                                }
                                return nodes.ToArray();
                            }
                            else
                            {
                                TreeNode[] nodes = new TreeNode[TagClasses.Length];
                                for (int x = 0; x < TagClasses.Length; x++)
                                {
                                    nodes[x] = new TreeNode(TagClasses[x], 1, 1);
                                    for (int y = 0; y < TagsList.Length; y++)
                                    {
                                        if (TagsList[y].Type == TagClasses[x])
                                            nodes[x].Nodes.Add(string.Empty, TagsList[y].Path, 2, 2);
                                    }
                                }
                                return nodes;
                            }
                        }
                    case SortTagsBy.PathHierarchy:
                        {
                            List<TreeNode> nodes = new List<TreeNode>(TagsList.Length);
                            for (int x = 0; x < TagsList.Length; x++)
                            {
                                if (string.IsNullOrEmpty(TagsList[x].Path) || !TagsList[x].Path.Contains("\\"))
                                {
                                    nodes.Add(new TreeNode(string.Format("{0}.{1}", TagsList[x].Path, TagsList[x].Type), 2, 2));
                                }
                                else
                                {
                                    string[] hier = TagsList[x].Path.Split('\\');
                                    TreeNode curNode = null;

                                    // Check first if node already exists.
                                    for (int y = 0; y < nodes.Count; y++)
                                    {
                                        if (nodes[y].Text == hier[0])
                                        {
                                            curNode = nodes[y];
                                            break;
                                        }
                                    }

                                    // If it didn't, add it.
                                    if (curNode == null)
                                    {
                                        curNode = new TreeNode(hier[0], 1, 1);
                                        nodes.Add(curNode);
                                    }

                                    // Loop through remaining pieces
                                    for (int y = 1; y < hier.Length - 1; y++)
                                    {
                                        bool exists = false;
                                        // Check first if node already exists.
                                        foreach (TreeNode tn in curNode.Nodes)
                                        {
                                            if (tn.Text == hier[y])
                                            {
                                                curNode = tn;
                                                exists = true;
                                                break;
                                            }
                                        }
                                        // If it didn't, add it.
                                        if (!exists) curNode = curNode.Nodes.Add(string.Empty, hier[y], 1, 1);
                                    }
                                    // Finally add the tags node.
                                    curNode.Nodes.Add(string.Empty, string.Format("{0}.{1}", hier[hier.Length - 1], TagsList[x].Type), 2, 2);
                                }
                            }
                            return nodes.ToArray();
                        }
                    //*
                    case SortTagsBy.ObjectHierarchy:
                        {
                            List<TreeNode> nodes = new List<TreeNode>(GrandParentClasses.Length);
                            for (int x = 0; x < GrandParentClasses.Length; x++)
                            {
                                bool cont = true;
                                for (int y = 0; y < nodes.Count; y++)
                                {
                                    if (nodes[y].Text == GrandParentClasses[x])
                                    {
                                        cont = false;
                                        break;
                                    }
                                }
                                if (!cont) continue;

                                TreeNode grandParentNode = new TreeNode(GrandParentClasses[x], 1, 1);
                                nodes.Add(grandParentNode);

                                for (int y = 0; y < ParentClasses.Length; y++)
                                {
                                    if (GrandParentClasses[y] == GrandParentClasses[x])
                                    {
                                        TreeNode parentNode = null;
                                        cont = true;
                                        for (int z = 0; z < grandParentNode.Nodes.Count; z++)
                                        {
                                            if (grandParentNode.Nodes[z].Text == ParentClasses[y])
                                            {
                                                parentNode = grandParentNode.Nodes[z];
                                                cont = false;
                                                break;
                                            }
                                        }
                                        if (cont)
                                        {
                                            parentNode = new TreeNode(ParentClasses[y], 1, 1);
                                            grandParentNode.Nodes.Add(parentNode);
                                        }

                                        for (int z = 0; z < TagClasses.Length; z++)
                                        {
                                            if (ParentClasses[z] == ParentClasses[y])
                                            {
                                                cont = true;
                                                for (int i = 0; i < parentNode.Nodes.Count; i++)
                                                {
                                                    if (parentNode.Nodes[i].Text == TagClasses[z])
                                                    {
                                                        cont = false;
                                                        break;
                                                    }
                                                }
                                                if (!cont) continue;

                                                TreeNode classNode = new TreeNode(TagClasses[z], 1, 1);
                                                parentNode.Nodes.Add(classNode);


                                                for (int i = 0; i < TagsList.Length; i++)
                                                {
                                                    if (TagsList[i].Type == TagClasses[z])
                                                        classNode.Nodes.Add(string.Empty, TagsList[i].Path, 2, 2);
                                                }
                                            }
                                        }
                                    }
                                }
                            }

                            if (removeEmptyClasses) // The lengths I go to..
                            {
                                for (int x = 0; x < nodes.Count; x++)
                                {
                                    for (int y = 0; y < nodes[x].Nodes.Count; y++)
                                    {
                                        for (int z = 0; z < nodes[x].Nodes[y].Nodes.Count; z++)
                                        {
                                            if (nodes[x].Nodes[y].Nodes[z].Nodes.Count == 0)
                                            {
                                                // Remove class node
                                                nodes[x].Nodes[y].Nodes.RemoveAt(z);
                                                z--;
                                            }
                                        }
                                        if (nodes[x].Nodes[y].Nodes.Count == 0)
                                        {
                                            // Remove parent node
                                            nodes[x].Nodes.RemoveAt(y);
                                            y--;
                                        }
                                    }
                                    if (nodes[x].Nodes.Count == 0)
                                    {
                                        // Remove grandparent node
                                        nodes.RemoveAt(x);
                                        x--;
                                    }
                                }
                            }

                            return nodes.ToArray();
                        }
                    //star/
                    default:
                        {
                            return null;
                        }
                }
            }
            else // Load specified type
            {
                TreeNode[] tnList = new TreeNode[1];
                tnList[0] = new TreeNode(type);

                for (int y = 0; y < TagsList.Length; y++)
                {
                    if (TagsList[y].Type == type)
                        tnList[0].Nodes.Add(TagsList[y].Path);
                }
                return tnList;
            }
        }

        public object[] ParseForList(string type)
        {
            if (type == string.Empty)
            {
                object[] objList = new object[Map.Index.TagCount];
                for (int x = 0; x < Map.Index.TagCount; x++)
                {
                    objList[x] = TagsList[x].Path;
                }
                return objList;
            }
            else
            {
                List<object> objList = new List<object>();
                for (int x = 0; x < Map.Index.TagCount; x++)
                {
                    if (TagsList[x].Type == type)
                        objList.Add(TagsList[x].Path);
                }
                return objList.ToArray();
            }
        }

        /// <summary>
        /// Writes the tag at the specified index to the map file.
        /// </summary>
        /// <param name="index">The specified index. </param>
        public void Write(int index)
        {
            BitWriter bw = new BitWriter(Map.FileName, Endian.Big);
            this.Write(bw, index);
            bw.Close();
        }

        /// <summary>
        /// Writes the tag at the specified index to the map file
        /// using the supplied BitWriter.
        /// </summary>
        /// <param name="bw">The supplied BitWriter. </param>
        /// <param name="index">The specified index. </param>
        public void Write(BitWriter bw, int index)
        {
            // TODO.
        }

        /// <summary>
        /// Writes all tags stored in the class to the map file.
        /// </summary>
        public void WriteAll()
        {
            BitWriter bw = new BitWriter(Map.FileName, Endian.Big);
            this.WriteAll(bw);
            bw.Close();
        }

        /// <summary>
        /// Writes all tags stored in the class to the map file
        /// using the supplied BitWriter.
        /// </summary>
        /// <param name="bw">The supplied BitWriter. </param>
        public void WriteAll(BitWriter bw)
        {
            for (int x = 0; x < Map.Index.TagCount; x++)
            {
                this.Write(bw, x);
            }
        }
    }

    /// <summary>
    /// Stores data relevant to the Tag structure.
    /// </summary>
    public struct Tag
    {
        #region Instance Data

        private int __identity__;
        private string __type__;
        private string __path__;
        private int __pathOffset__;
        private int __pathLength__;
        private int __rawMetaOffset__;
        private int __metaOffset__;
        private int __metaSize__;

        #endregion

        #region Accessors

        ///<summary>
        /// Identity
        ///</summary>
        public int Identity { get { return __identity__; } set { __identity__ = value; } }
        ///<summary>
        /// Type
        ///</summary>
        public string Type { get { return __type__; } set { if (value.Length <= 4) __type__ = value; } }
        ///<summary>
        /// Path
        ///</summary>
        public string Path { get { return __path__; } set { __path__ = value; } }
        ///<summary>
        /// PathOffset
        ///</summary>
        public int PathOffset { get { return __pathOffset__; } set { __pathOffset__ = value; } }
        ///<summary>
        /// PathLength
        ///</summary>
        public int PathLength { get { return __pathLength__; } set { __pathLength__ = value; } }
        ///<summary>
        /// RawMetaOffset
        ///</summary>
        public int RawMetaOffset { get { return __rawMetaOffset__; } set { __rawMetaOffset__ = value; } }
        ///<summary>
        /// MetaOffset
        ///</summary>
        public int MetaOffset { get { return __metaOffset__; } set { __metaOffset__ = value; } }
        ///<summary>
        /// MetaSize
        ///</summary>
        public int MetaSize { get { return __metaSize__; } set { __metaSize__ = value; } }

        #endregion

        /// <summary>
        /// Creates an exact replica of the current H3UtilityToolkit.Essentials.Tag.
        /// </summary>
        public Tag Clone()
        {
            Tag tag = new Tag();
            tag.Identity = this.Identity;
            tag.Type = this.Type;
            tag.Path = this.Path;
            tag.PathOffset = this.PathOffset;
            tag.PathLength = this.PathLength;
            tag.RawMetaOffset = this.RawMetaOffset;
            tag.MetaOffset = this.MetaOffset;
            tag.MetaSize = this.MetaSize;
            return tag;
        }
    }

    /// <summary>
    /// Sort mode for the Tags class.
    /// </summary>
    public enum SortTagsBy
    {
        TagClass,
        PathHierarchy,
        ObjectHierarchy
    }
}