using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.IO;
using System.Extended;
using H3UtilityToolkit.Essentials;
using H3UtilityToolkit.Misc;
using H3UtilityToolkit.PluginStuffs;

namespace H3UtilityToolkit
{
    /// <summary>
    /// Object for reading, storing, and editing anything and everything to do with the map file.
    /// </summary>
    public class Map
    {
        #region Instance Data

        private string __fileName__;
        private bool __loaded__;
        private Header __header__;
        private Index __index__;
        private Tags __tags__;
        private Strings __strings__;
        private Analyser __analyser__;
        private MiscUtilities __misc__;
        private Preferences __prefs__;
        private PluginManager __pluginManager__;

        #endregion

        #region Accessors

        ///<summary>
        /// FileName
        ///</summary>
        public string FileName { get { return __fileName__; } set { __fileName__ = value; } }
        /// <summary>
        /// Loaded
        /// </summary>
        public bool Loaded { get { return __loaded__; } set { __loaded__ = value; } }
        ///<summary>
        /// Header
        ///</summary>
        public Header Header { get { return __header__; } set { __header__ = value; } }
        ///<summary>
        /// Index
        ///</summary>
        public Index Index { get { return __index__; } set { __index__ = value; } }
        ///<summary>
        /// Tags
        ///</summary>
        public Tags Tags { get { return __tags__; } set { __tags__ = value; } }
        ///<summary>
        /// Strings
        ///</summary>
        public Strings Strings { get { return __strings__; } set { __strings__ = value; } }
        /// <summary>
        /// Analyser
        /// </summary>
        public Analyser Analyser { get { return __analyser__; } set { __analyser__ = value; } }
        /// <summary>
        /// Misc
        /// </summary>
        public MiscUtilities Misc { get { return __misc__; } set { __misc__ = value; } }
        /// <summary>
        /// Prefs
        /// </summary>
        public Preferences Prefs { get { return __prefs__; } set { __prefs__ = value; } }
        /// <summary>
        /// PluginManager
        /// </summary>
        public PluginManager PluginManager { get { return __pluginManager__; } set { __pluginManager__ = value; } }

        #endregion

        /// <summary>
        /// Initialises a new instance of the Map class.
        /// </summary>
        public Map()
        {
            this.Analyser = new Analyser(this);
            this.Prefs = new Preferences(this);
            this.Misc = new MiscUtilities(this);
            this.PluginManager = new PluginManager(this);
        }

        /// <summary>
        /// Initialises a new instance of the Map class, and begins reading 
        /// straight away from the supplied file name.
        /// </summary>
        /// <param name="filename">The supplied file name. </param>
        public Map(string fileName)
        {
            this.Analyser = new Analyser(this);
            this.Prefs = new Preferences(this);
            this.Misc = new MiscUtilities(this);
            this.PluginManager = new PluginManager(this);
            this.Load(fileName);
        }

        /// <summary>
        /// Initialises a new instance of the Map class, and begins reading 
        /// straight away with the supplied BitReader.
        /// </summary>
        /// <param name="br">The supplied BitReader. </param>
        public Map(BitReader br)
        {
            this.Analyser = new Analyser(this);
            this.Prefs = new Preferences(this);
            this.Misc = new MiscUtilities(this);
            this.PluginManager = new PluginManager(this);
            this.Load(br);
        }

        /// <summary>
        /// Reads data from the supplied file name and stores it for access later.
        /// </summary>
        /// <param name="filename">The supplied file name. </param>
        public void Load(string fileName)
        {
            BitReader br = new BitReader(fileName, Endian.Big);
            this.Load(br);
            br.Close();
        }

        /// <summary>
        /// Reads data with the supplied BitReader and stores it for access later.
        /// </summary>
        /// <param name="br">The supplied BitReader. </param>
        public void Load(BitReader br)
        {
            this.Loaded = false;
            this.FileName = (br.BaseStream as System.IO.FileStream).Name;

            this.Header = new Header(this, br);
            this.Index = new Index(this, br);
            this.Tags = new Tags(this, br);
            this.Strings = new Strings(this, br);

            this.Loaded = true;
        }

        /// <summary>
        /// Re-calculates the map's XOR checksum. 
        /// </summary>
        public void Resign()
        {
            // Use Litte Endian as it's quicker, and produces same result.
            BitReader br = new BitReader(FileName, Endian.Little);
            BitWriter bw = new BitWriter(br.BaseStream, Endian.Little);

            int sig = 0;
            long length = br.BaseStream.Length;

            br.BaseStream.Position = Header.Length;

            for (long x = Header.Length; x < length; x += 4)
                sig ^= br.ReadInt32();

            bw.BaseStream.Position = 708;
            bw.Write(sig);

            br.Endian = Endian.Big;
            br.BaseStream.Position = 708;
            Header.Checksum = br.ReadInt32();

            bw.Close();
            br.Close();
        }
    }
}