using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.IO;
using System.Extended;
using H3UtilityToolkit;
using H3UtilityToolkit.Essentials;
using H3UtilityToolkit.Structures;

namespace H3UtilityToolkit.Misc
{
    /// <summary>
    /// Parses map file into visual representation.
    /// </summary>
    public class Analyser
    {
        public Map Map;

        /// <summary>
        /// Initialises a new instance of the Analyser class based on the
        /// supplied Map reference.
        /// </summary>
        /// <param name="map">The supplied Map reference. </param>
        public Analyser(Map map)
        {
            this.Map = map;
        }

        /// <summary>
        /// Analyses the map file, and produces a visual representation
        /// constrained to the boundaries of the specified width.
        /// </summary>
        /// <param name="ParentWidth">The specified width. </param>
        public Control[] Analyse(int parentWidth)
        {
            BitReader br = new BitReader(Map.FileName, Endian.Big);
            Control[] controls = Analyse(br, parentWidth);
            br.Close();
            return controls;
        }

        /// <summary>
        /// Analyses the map file with the supplied BitReader, 
        /// and produces a visual representation constrained 
        /// to the boundaries of the specified width.
        /// </summary>
        /// <param name="br">The supplied BitReader. </param>
        /// <param name="ParentWidth">The specified width. </param>
        public Control[] Analyse(BitReader br, int parentWidth)
        {
            bool ADD_REFLEXIVES = false; // If true, slows the function waaaaaaay down.

            List<AnalyzePiece> aps = new List<AnalyzePiece>();
            int curOffset;
            int curSize;
            int curPadding;

            #region Get

            //--[Header]
            curOffset = 0;
            curSize = Header.Length;
            aps.Add(new AnalyzePiece("Header", curOffset, curSize, 0, AnalyzeType.Other, Color.DarkGoldenrod));

            //--[Index]
            if (Map.Header.IndexOffset >= Header.Length)
            {
                curOffset = Map.Header.IndexOffset;
                curSize = Index.Length;
                aps.Add(new AnalyzePiece("IndexHeader", curOffset, curSize, 0, AnalyzeType.Other, Color.Gray));
            }

            if (Map.Index.TagInfoHeaderOffset >= Header.Length)
            {
                curOffset = Map.Index.TagInfoHeaderOffset;
                curSize = Map.Index.TagInfoHeaderCount * 8;
                aps.Add(new AnalyzePiece("TagInfoHeader Index", curOffset, curSize, 0, AnalyzeType.Other, Color.DarkSlateGray));
            }

            if (Map.Index.TagClassIndexOffset >= Header.Length)
            {
                curOffset = Map.Index.TagClassIndexOffset;
                curSize = Map.Index.TagClassCount * 16;
                aps.Add(new AnalyzePiece("TagClass Index", curOffset, curSize, 0, AnalyzeType.Other, Color.DarkSlateGray));
            }

            if (Map.Index.TagInfoOffset >= Header.Length)
            {
                curOffset = Map.Index.TagInfoOffset;
                curSize = Map.Index.TagCount * 8;
                aps.Add(new AnalyzePiece("TagInfo", curOffset, curSize, 0, AnalyzeType.Other, Color.LightSlateGray));
            }

            if (Map.Header.StringTableIndexOffset >= Header.Length)
            {
                //--[Strings]
                curSize = Map.Header.StringTableCount * 4;
                curPadding = (4096 - (curSize % 4096)) % 4096;
                aps.Add(new AnalyzePiece("StringID Index", Map.Header.StringTableIndexOffset,
                    curSize, curPadding, AnalyzeType.Other, Color.LightSeaGreen));
            }

            if (Map.Header.StringTableOffset >= Header.Length)
            {
                curPadding = (4096 - (Map.Header.StringTableSize % 4096)) % 4096;
                aps.Add(new AnalyzePiece("StringID Table", Map.Header.StringTableOffset,
                    Map.Header.StringTableSize, curPadding, AnalyzeType.Other, Color.LightGoldenrodYellow));
            }

            if (Map.Header.FileTableIndexOffset >= Header.Length)
            {
                //--[File]
                curSize = Map.Header.FileTableCount * 4;
                curPadding = (4096 - (curSize % 4096)) % 4096;
                aps.Add(new AnalyzePiece("File Index", Map.Header.FileTableIndexOffset,
                    curSize, curPadding, AnalyzeType.Other, Color.Thistle));
            }

            if (Map.Header.FileTableOffset >= Header.Length) // Not protected
            {
                curPadding = (4096 - (Map.Header.FileTableSize % 4096)) % 4096;
                aps.Add(new AnalyzePiece("File Table", Map.Header.FileTableOffset,
                    Map.Header.FileTableSize, curPadding, AnalyzeType.Other, Color.Goldenrod));
            }

            //--[Locales]
            int firstLocaleOffset = Map.Header.Filesize;
            int matgIndx = Map.Tags.IndexOf("matg", "globals\\globals");
            if (matgIndx != -1)
            {
                string[] langs = Enum.GetNames(typeof(Matg.Locale));

                const string wordIndex = " Index";
                const string wordTable = " Table";
                br.BaseStream.Position = Map.Tags.TagsList[matgIndx].MetaOffset + 452;
                for (int x = 0; x < langs.Length; x++)
                {
                    int langTableCount = br.ReadInt32();
                    int langTableSize = br.ReadInt32();
                    int langTableIndexOffset = br.ReadInt32() + Map.Header.LocaleTableAddrMod;
                    int langTableOffset = br.ReadInt32() + Map.Header.LocaleTableAddrMod;
                    int langTableIndexSize = langTableCount * 8;

                    if (x == 0) firstLocaleOffset = langTableIndexOffset;

                    curPadding = (4096 - (langTableIndexSize % 4096)) % 4096;
                    aps.Add(new AnalyzePiece(langs[x] + wordIndex, langTableIndexOffset,
                        langTableIndexSize, curPadding, AnalyzeType.Other,
                        Color.Green));

                    curPadding = (4096 - (langTableSize % 4096)) % 4096;
                    aps.Add(new AnalyzePiece(langs[x] + wordTable, langTableOffset,
                        langTableSize, curPadding, AnalyzeType.Other,
                        Color.DarkGreen));

                    br.BaseStream.Position += 52;
                }
            }

            //--[Tags]
            const string wordMeta = "Meta";
            for (int x = 0; x < Map.Index.TagCount; x++)
            {
                string tagFilename;
                try { tagFilename = Path.GetFileName(Map.Tags.TagsList[x].Path); }
                catch { tagFilename = "Error Processing Name"; }

                aps.Add(new AnalyzePiece(tagFilename + wordMeta, Map.Tags.TagsList[x].MetaOffset,
                        Map.Tags.TagsList[x].MetaSize, 0, AnalyzeType.Meta, Color.RoyalBlue));
            }

            //--[Assets]
            int zoneIndx = Map.Tags.IndexOf("zone", "there they are all standing in a row");
            if (zoneIndx != -1)
            {
                br.BaseStream.Position = Map.Tags.TagsList[zoneIndx].MetaOffset + 88;

                int cc = br.ReadInt32();
                int tr = br.ReadInt32() - Map.Header.MapMagic;

                aps.Add(new AnalyzePiece("Zone Raw Table", tr,
                    cc * 64, 0, AnalyzeType.Other, Color.LightGray));
            }

            if (ADD_REFLEXIVES)
            {
                //--[Reflexives]
                for (int x = 0; x < Map.Index.TagCount; x++)
                {
                    br.BaseStream.Position = Map.Tags.TagsList[x].MetaOffset;
                    for (int y = 0; y < Map.Tags.TagsList[x].MetaSize; y += 4)
                    {
                        if (br.BaseStream.Position == firstLocaleOffset) break;

                        int cc = br.ReadInt32();
                        int ptr = br.ReadInt32();
                        int tr = ptr - Map.Header.MapMagic;

                        if (cc >= 0 && ptr < 0 && tr >= Map.Header.VirtSegmentStart && tr < firstLocaleOffset)
                        {
                            aps.Add(new AnalyzePiece(string.Format("R{0}{1}", x, y), tr,
                                -1, 0, AnalyzeType.Reflexive, Color.Khaki));
                        }
                        else br.BaseStream.Position -= 4;
                    }
                }
            }

            #endregion

            #region Remove Duplicates

            for (int x = aps.Count - 1; x >= 0; x--)
            {
                for (int y = 0; y < aps.Count; y++)
                {
                    if (x != y &&
                        aps[x].Type == aps[y].Type &&
                        aps[x].Offset == aps[y].Offset &&
                        aps[x].Size == aps[y].Size)
                    {
                        aps.RemoveAt(x);
                        break;
                    }
                }
            }

            #endregion

            #region Sort

            List<AnalyzePiece> tmpList = new List<AnalyzePiece>();
            while (aps.Count != 0)
            {
                int sIndx = 0;
                for (int x = 1; x < aps.Count; x++)
                {
                    if (aps[x].Offset < aps[sIndx].Offset) sIndx = x;
                }
                tmpList.Add(aps[sIndx]);
                aps.RemoveAt(sIndx);
            }
            aps = tmpList;

            #endregion

            if (ADD_REFLEXIVES)
            {
                #region Get Reflexive Sizes

                for (int x = aps.Count - 1; x > 0; x--)
                {
                    if (aps[x - 1].Type == AnalyzeType.Reflexive)
                    {
                        aps[x - 1].Size = aps[x].Offset - aps[x].Offset;
                    }
                }

                #endregion
            }

            #region Group Meta

            AnalyzeType at = AnalyzeType.Meta;
            bool onSomething = false;
            for (int y = aps.Count - 1; y >= 0; y--)
            {
                if (aps[y].Type == at)
                {
                    if (onSomething)
                    {
                        aps[y].Size += aps[y + 1].Size + aps[y + 1].Padding;
                        aps[y].ItemCount += 1;
                        aps.RemoveAt(y + 1);
                    }
                    else
                    {
                        if (y != aps.Count - 1)
                        {
                            aps[y].Size = aps[y + 1].Offset - aps[y].Offset;
                            aps[y].Padding = 0;
                        }
                        onSomething = true;
                    }
                }
                else if (onSomething)
                {
                    aps[y + 1].Name = at.ToString();
                    onSomething = false;
                }
            }

            #endregion

            #region Group

            int atCount = Enum.GetNames(typeof(AnalyzeType)).Length;
            for (int x = 1; x != atCount; x++)
            {
                at = (AnalyzeType)x;

                // Already done this one.
                if (at == AnalyzeType.Meta) continue;

                onSomething = false;
                for (int y = aps.Count - 1; y >= 0; y--)
                {
                    if (aps[y].Type == at)
                    {
                        if (onSomething)
                        {
                            aps[y].Size += aps[y + 1].Size + aps[y + 1].Padding;
                            aps[y].ItemCount += 1;
                            aps.RemoveAt(y + 1);
                        }
                        else
                        {
                            onSomething = true;
                        }
                    }
                    else if (onSomething)
                    {
                        aps[y + 1].Name = at.ToString();
                        onSomething = false;
                    }
                }
            }

            #endregion

            #region Pad Groups

            for (int x = 0; x < aps.Count; x++)
            {
                if (aps[x].ItemCount == 1) continue;

                if (aps[x].Padding != 0)
                {
                    aps[x].Size += aps[x].Padding;
                    aps[x].Padding = 0;
                }
            }

            #endregion

            #region Identify Overwritten Data

            for (int x = 0; x < aps.Count; x++)
            {
                if (aps[x].Size == 0) continue;
                for (int y = 0; y < aps.Count; y++)
                {
                    if (x != y &&
                        aps[x].Offset >= aps[y].Offset &&
                        aps[x].Offset < (aps[y].Offset + aps[y].Size + aps[y].Padding))
                    {
                        string yName = aps[y].Name;
                        if (aps[y].Colour == Color.Orchid)
                        {
                            yName = yName.Remove(yName.IndexOf(" - In "));
                        }

                        aps[x].Colour = Color.Orchid;
                        aps[x].Name += " - In " + yName + ". NTS " +
                            ((aps[y].Offset + aps[y].Size + aps[y].Padding) - aps[x].Offset);
                    }
                }
            }

            #endregion

            #region Identify Gaps

            for (int x = aps.Count - 1; x > 0; x--)
            {
                if ((aps[x - 1].Offset + aps[x - 1].Size + aps[x - 1].Padding) < aps[x].Offset)
                {
                    aps.Insert(x, new AnalyzePiece("Undefined",
                        aps[x - 1].Offset + aps[x - 1].Size + aps[x - 1].Padding,
                        aps[x].Offset - (aps[x - 1].Offset + aps[x - 1].Size + aps[x - 1].Padding),
                        0, AnalyzeType.Other, Color.Teal));
                }
            }
            int min1 = aps.Count - 1;
            if ((aps[min1].Offset + aps[min1].Size + aps[min1].Padding) < Map.Header.Filesize)
            {
                aps.Add(new AnalyzePiece("Undefined",
                    aps[min1].Offset + aps[min1].Size + aps[min1].Padding,
                    Map.Header.Filesize - (aps[min1].Offset + aps[min1].Size + aps[min1].Padding),
                    0, AnalyzeType.Other, Color.Teal));
            }

            #endregion

            #region Work Out Percentages

            for (int x = 0; x < aps.Count; x++)
            {
                aps[x].CalculatePercent(Map.Header.Filesize);
            }

            #endregion

            #region Display

            Control[] controls = new Control[aps.Count];
            int ypos = 6;

            for (int x = 0; x < aps.Count; x++)
            {
                AnalyzedPiece ap = new AnalyzedPiece(Map.FileName, aps[x]);
                ap.Left = 0;
                ap.Top = ypos;
                ap.Width = parentWidth - 12;
                controls[x] = ap;

                ypos += ap.Height + 2;
            }
            return controls;

            #endregion
        }
    }

    /// <summary>
    /// The Type of Item that has been analysed.
    /// </summary>
    public enum AnalyzeType
    {
        Other,
        Meta,
        Reflexive
    }

    /// <summary>
    /// An item that stores data on an object in the map file.
    /// </summary>
    public class AnalyzePiece
    {
        public string Name;
        public int Offset;
        public int Size;
        public int Padding;
        public float Percent;
        public AnalyzeType Type;
        public Color Colour;
        public int ItemCount;

        public AnalyzePiece(string Name, int Offset, int Size, int Padding, AnalyzeType Type, Color Colour)
        {
            this.Name = Name;
            this.Offset = Offset;
            this.Size = Size;
            this.Padding = Padding;
            this.Percent = 0;
            this.Type = Type;
            this.Colour = Colour;
            this.ItemCount = 1;
        }

        public AnalyzePiece(string Name, int Offset, int Size, int Padding, float Percent, AnalyzeType Type, Color Colour)
        {
            this.Name = Name;
            this.Offset = Offset;
            this.Size = Size;
            this.Padding = Padding;
            this.Percent = Percent;
            this.Type = Type;
            this.Colour = Colour;
            this.ItemCount = 1;
        }

        public void CalculatePercent(int mapLength)
        {
            float one = (float)(Size + Padding) / (float)mapLength;
            Percent = one * 100f;
            Percent = Mathematics.Round(Percent, 3);
        }
    }
}
