using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.IO;
using System.Extended;
using H3UtilityToolkit.PluginStuffs;

namespace H3UtilityToolkit.Misc
{
    /// <summary>
    /// For storing user preferences.
    /// </summary>
    public class Preferences
    {
        public const int MaxRecentFiles = 5;
        public static List<RecentFile> RecentMaps;
        public static List<RecentFile> RecentImages;
        public static List<RecentFile> RecentMapInfos;

        #region Instance Data

        #region Readonly

        /// <summary>
        /// FileName
        /// </summary>
        public static readonly string FileName;

        #endregion

        protected Map __map__;
        private static string __mainMenuFileName__;
        private static string __mPSharedFilename__;
        private static string __sPSharedFileName__;
        private bool __mainMenuQueried__;
        private bool __mPSharedQueried__;
        private bool __sPSharedQueried__;

        #endregion

        #region Accessors

        ///<summary>
        /// Map
        ///</summary>
        protected Map Map { get { return __map__; } set { __map__ = value; } }
        ///<summary>
        /// MainMenuFileName
        ///</summary>
        public static string MainMenuFileName { get { return __mainMenuFileName__; } set { __mainMenuFileName__ = value; } }
        ///<summary>
        /// MPSharedFilename
        ///</summary>
        public static string MPSharedFilename { get { return __mPSharedFilename__; } set { __mPSharedFilename__ = value; } }
        ///<summary>
        /// SPSharedFileName
        ///</summary>
        public static string SPSharedFileName { get { return __sPSharedFileName__; } set { __sPSharedFileName__ = value; } }
        ///<summary>
        /// MainMenuQueried
        ///</summary>
        public bool MainMenuQueried { get { return __mainMenuQueried__; } set { __mainMenuQueried__ = value; } }
        ///<summary>
        /// MPSharedQueried
        ///</summary>
        public bool MPSharedQueried { get { return __mPSharedQueried__; } set { __mPSharedQueried__ = value; } }
        ///<summary>
        /// SPSharedQueried
        ///</summary>
        public bool SPSharedQueried { get { return __sPSharedQueried__; } set { __sPSharedQueried__ = value; } }

        #endregion

        /// <summary>
        /// Initialises a new instance of the Preferences class based on the
        /// supplied Map reference.
        /// </summary>
        /// <param name="map">The supplied Map reference. </param>
        public Preferences(Map map)
        {
            this.Map = map;
        }

        /// <summary>
        /// Static Startup.
        /// </summary>
        static Preferences()
        {
            FileName = Application.StartupPath + @"\User.config";

#if DEBUG
            FileName = Path.GetDirectoryName(Application.StartupPath) + @"\Release\User.config";
#endif

            RecentMaps = new List<RecentFile>(MaxRecentFiles);
            RecentImages = new List<RecentFile>(MaxRecentFiles);
            RecentMapInfos = new List<RecentFile>(MaxRecentFiles);
        }

        /// <summary>
        /// Reads the preferences file and uses it to populate
        /// current preference members. If no file found members
        /// are default.
        /// </summary>
        public static void Read()
        {
            if (File.Exists(FileName))
            {
                BitReader br = new BitReader(FileName, Endian.Little);
                try
                {
                    MainMenuFileName = br.ReadCString();
                    MPSharedFilename = br.ReadCString();
                    SPSharedFileName = br.ReadCString();

                    byte count = br.ReadByte(); count++;
                    for (byte x = 1; x < count; x++)
                    {
                        string fileName = br.ReadCString();
                        if (File.Exists(fileName))
                            RecentMaps.Add(new RecentFile((byte)(count - x), fileName));
                    }
                    if (RecentMaps.Count > MaxRecentFiles)
                        RecentMaps.RemoveRange(MaxRecentFiles, RecentMaps.Count - MaxRecentFiles);

                    count = br.ReadByte(); count++;
                    for (byte x = 1; x < count; x++)
                    {
                        string filePath = br.ReadCString();
                        if (Directory.Exists(filePath))
                            RecentImages.Add(new RecentFile((byte)(count - x), filePath));
                    }
                    if (RecentImages.Count > MaxRecentFiles)
                        RecentImages.RemoveRange(MaxRecentFiles, RecentImages.Count - MaxRecentFiles);

                    count = br.ReadByte(); count++;
                    for (byte x = 1; x < count; x++)
                    {
                        string filePath = br.ReadCString();
                        if (Directory.Exists(filePath))
                            RecentMapInfos.Add(new RecentFile((byte)(count - x), filePath));
                    }
                    if (RecentMapInfos.Count > MaxRecentFiles)
                        RecentMapInfos.RemoveRange(MaxRecentFiles, RecentMapInfos.Count - MaxRecentFiles);

                    PluginManager.VisibleMask = NodeType.None;
                    string[] dataTypes = Enum.GetNames(typeof(NodeType));
                    for (int x = 0; x < dataTypes.Length; x++)
                    {
                        NodeType nt = (NodeType)Enum.Parse(typeof(NodeType), dataTypes[x]);
                        if (nt != NodeType.None && nt != NodeType.All && nt != NodeType.BoxUsers)
                        {
                            if (br.ReadBoolean())
                            {
                                PluginManager.VisibleMask |= nt;
                            }
                        }
                    }

                    PluginManager.ReflexiveBaseColour = Color.FromArgb(br.ReadInt32());
                    PluginManager.ReflexiveRedShift = br.ReadInt32();
                    PluginManager.ReflexiveGreenShift = br.ReadInt32();
                    PluginManager.ReflexiveBlueShift = br.ReadInt32();

                }
                catch { }

                br.Close();
            }
            else
            {
                // Defaults
                MainMenuFileName = string.Empty;
                MPSharedFilename = string.Empty;
                SPSharedFileName = string.Empty;

                PluginManager.VisibleMask = NodeType.All;

                PluginManager.ReflexiveBaseColour = Color.FromArgb(-9982777); // Blue
                PluginManager.ReflexiveRedShift = 61;
                PluginManager.ReflexiveGreenShift = 39;
                PluginManager.ReflexiveBlueShift = 15;
            }
        }

        /// <summary>
        /// Writes preference members to preference file.
        /// </summary>
        public static void Write()
        {
            try
            {
                BitWriter bw = new BitWriter(new FileStream(FileName, FileMode.Create,
                    FileAccess.ReadWrite, FileShare.ReadWrite), Endian.Little);

                byte byteNull = 0;

                if (!string.IsNullOrEmpty(MainMenuFileName)) bw.Write(MainMenuFileName);
                bw.Write(byteNull);
                if (!string.IsNullOrEmpty(MPSharedFilename)) bw.Write(MPSharedFilename);
                bw.Write(byteNull);
                if (!string.IsNullOrEmpty(SPSharedFileName)) bw.Write(SPSharedFileName);
                bw.Write(byteNull);

                bw.Write((byte)RecentMaps.Count);
                for (int x = 0; x < RecentMaps.Count; x++)
                {
                    bw.Write(RecentMaps[x].FileName);
                    bw.Write(byteNull);
                }

                bw.Write((byte)RecentImages.Count);
                for (int x = 0; x < RecentImages.Count; x++)
                {
                    bw.Write(RecentImages[x].FileName);
                    bw.Write(byteNull);
                }

                bw.Write((byte)RecentMapInfos.Count);
                for (int x = 0; x < RecentMapInfos.Count; x++)
                {
                    bw.Write(RecentMapInfos[x].FileName);
                    bw.Write(byteNull);
                }

                string[] dataTypes = Enum.GetNames(typeof(NodeType));
                for (int x = 0; x < dataTypes.Length; x++)
                {
                    NodeType nt = (NodeType)Enum.Parse(typeof(NodeType), dataTypes[x]);
                    if (nt != NodeType.None && nt != NodeType.All && nt != NodeType.BoxUsers)
                    {
                        bw.Write(PluginManager.IsVisible(nt));
                    }
                }

                bw.Write(PluginManager.ReflexiveBaseColour.ToArgb());
                bw.Write(PluginManager.ReflexiveRedShift);
                bw.Write(PluginManager.ReflexiveGreenShift);
                bw.Write(PluginManager.ReflexiveBlueShift);

                bw.Close();
            }
            catch { }
        }

        /// <summary>
        /// Resets current external map query states.
        /// </summary>
        public void ResetQueryStates()
        {
            MainMenuQueried = false;
            MPSharedQueried = false;
            SPSharedQueried = false;
        }

        /// <summary>
        /// Querys for the specified map if map query state
        /// has not been disabled by user actions.
        /// </summary>
        /// <param name="rawLoc">The specified map. </param>
        public string QueryForExternalMap(AssetLocation assetLoc)
        {
            switch (assetLoc)
            {
                case AssetLocation.Internal:
                    return Map.FileName;

                case AssetLocation.MainMenu:
                    if (MainMenuQueried) return string.Empty;
                    break;

                case AssetLocation.MPShared:
                    if (MPSharedQueried) return string.Empty;
                    break;

                case AssetLocation.SPShared:
                    if (SPSharedQueried) return string.Empty;
                    break;
            }

            using (OpenFileDialog o = new OpenFileDialog())
            {
                o.Title = string.Format("\"{0}.map\" needed", Map.Misc.GetMapNameFromAssetLocation(assetLoc));
                o.Filter = "Map (*.map)|*.map|All Files (*.*)|*.*";

                if (o.ShowDialog() == DialogResult.Cancel)
                {
                    switch (assetLoc)
                    {
                        case AssetLocation.MainMenu:
                            MainMenuQueried = true;
                            break;

                        case AssetLocation.MPShared:
                            MPSharedQueried = true;
                            break;

                        case AssetLocation.SPShared:
                            SPSharedQueried = true;
                            break;
                    }
                    return string.Empty;
                }
                else
                {
                    switch (assetLoc)
                    {
                        case AssetLocation.MainMenu:
                            MainMenuFileName = o.FileName;
                            break;

                        case AssetLocation.MPShared:
                            MPSharedFilename = o.FileName;
                            break;

                        case AssetLocation.SPShared:
                            SPSharedFileName = o.FileName;
                            break;
                    }
                    return o.FileName;
                }
            }
        }

        /// <summary>
        /// Querys for the specified map regardless of map query state.
        /// </summary>
        /// <param name="rawLoc">The specified map. </param>
        public string QueryForExternalMapRegardless(AssetLocation assetLoc)
        {
            if (assetLoc == AssetLocation.Internal) return Map.FileName;
            using (OpenFileDialog o = new OpenFileDialog())
            {
                o.Title = string.Format("\"{0}.map\" needed", Map.Misc.GetMapNameFromAssetLocation(assetLoc));
                o.Filter = "Map (*.map)|*.map|All Files (*.*)|*.*";

                if (o.ShowDialog() != DialogResult.Cancel)
                {
                    switch (assetLoc)
                    {
                        case AssetLocation.MainMenu:
                            MainMenuFileName = o.FileName;
                            break;

                        case AssetLocation.MPShared:
                            MPSharedFilename = o.FileName;
                            break;

                        case AssetLocation.SPShared:
                            SPSharedFileName = o.FileName;
                            break;
                    }
                    return o.FileName;
                }
                else return string.Empty;
            }
        }

        /// <summary>
        /// Gets stream access for the specified map if map query state
        /// has not been disabled by user actions.
        /// </summary>
        /// <param name="rawLoc">The specified map. </param>
        public FileStream GetAssetStream(AssetLocation assetLoc)
        {
            string fileName = string.Empty;

            switch (assetLoc)
            {
                case AssetLocation.Internal:
                    fileName = Map.FileName;
                    break;

                case AssetLocation.MainMenu:
                    if (string.IsNullOrEmpty(MainMenuFileName)) MainMenuFileName = QueryForExternalMap(assetLoc);
                    fileName = MainMenuFileName;
                    break;

                case AssetLocation.MPShared:
                    if (string.IsNullOrEmpty(MPSharedFilename)) MPSharedFilename = QueryForExternalMap(assetLoc);
                    fileName = MPSharedFilename;
                    break;

                case AssetLocation.SPShared:
                    if (string.IsNullOrEmpty(SPSharedFileName)) SPSharedFileName = QueryForExternalMap(assetLoc);
                    fileName = SPSharedFileName;
                    break;
            }

            return (File.Exists(fileName) ? new FileStream(fileName, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite) : null);
        }

        /// <summary>
        /// Gets stream access for the specified map regardless of map query state.
        /// </summary>
        /// <param name="rawLoc">The specified map. </param>
        public FileStream GetAssetStreamRegardless(AssetLocation assetLoc)
        {
            string fileName = string.Empty;

            switch (assetLoc)
            {
                case AssetLocation.Internal:
                    fileName = Map.FileName;
                    break;

                case AssetLocation.MainMenu:
                    if (string.IsNullOrEmpty(MainMenuFileName)) MainMenuFileName = QueryForExternalMapRegardless(assetLoc);
                    fileName = MainMenuFileName;
                    break;

                case AssetLocation.MPShared:
                    if (string.IsNullOrEmpty(MPSharedFilename)) MPSharedFilename = QueryForExternalMapRegardless(assetLoc);
                    fileName = MPSharedFilename;
                    break;

                case AssetLocation.SPShared:
                    if (string.IsNullOrEmpty(SPSharedFileName)) SPSharedFileName = QueryForExternalMapRegardless(assetLoc);
                    fileName = SPSharedFileName;
                    break;
            }

            return (File.Exists(fileName) ? new FileStream(fileName, FileMode.Open,
                FileAccess.ReadWrite, FileShare.ReadWrite) : null);
        }

        public class RecentFile
        {
            #region Instance Data

            private string __fileName__;
            private ToolStripMenuItem __tool__;
            private byte __index__;

            #endregion

            #region Accessors

            /// <summary>
            /// FileName
            /// </summary>
            public string FileName { get { return __fileName__; } set { __fileName__ = value; } }
            /// <summary>
            /// Tool
            /// </summary>
            public ToolStripMenuItem Tool { get { return __tool__; } set { __tool__ = value; } }
            /// <summary>
            /// Index
            /// </summary>
            public byte Index { get { return __index__; } set { __index__ = value; Tool.Text = string.Format("{0} {1}", Index, FileName); } }

            #endregion

            public RecentFile(byte index, string fileName)
            {
                FileName = fileName;
                Tool = new ToolStripMenuItem(); Tool.Tag = this;
                Index = index;
            }
        }
    }
}