using System;
using System.Collections.Generic;
using System.Text;
using System.Extended;
using System.Drawing;
using System.Windows.Forms;
using H3UtilityToolkit.Essentials;

namespace H3UtilityToolkit.PluginStuffs
{
    [Flags]
    public enum NodeType
    {
        VisibleFalse = 0x1,
        Unknown = 0x2,
        Reflexive = 0x4,
        TagRef = 0x8,
        StringID = 0x10,
        Byte = 0x20,
        Short = 0x40,
        UShort = 0x80,
        Long = 0x100,
        ULong = 0x200,
        Quad = 0x400,
        UQuad = 0x800,
        Float = 0x1000,
        Double = 0x2000,
        Bitmask = 0x4000,
        Enum = 0x8000,
        String = 0x10000,
        Void = 0x20000,

        None = 0x40000,

        All = VisibleFalse | Unknown | Reflexive | TagRef | StringID | Byte | Short | UShort |
            Long | ULong | Quad | UQuad | Float | Double | Bitmask | Enum | String | Void,

        /// <summary>
        /// Controls include either TextBoxes or ComboBoxes.
        /// </summary>
        BoxUsers = Unknown | TagRef | StringID | Byte | Short | UShort | Long | ULong |
            Quad | UQuad | Float | Double | String | Void
    }

    public struct Revision
    {
        public string Author;
        public string Version;
        public string Comment;

        public Revision(string author, string version, string comment)
        {
            this.Author = author;
            this.Version = version;
            this.Comment = comment;
        }
    }

    public class BaseNode
    {
        #region Instance Data

        protected PluginManager __manager__;
        protected StructNode __parent__;
        protected NodeType __nodeType__;
        protected string __name__;
        protected int __offset__;
        protected int __size__;
        protected bool __visible__;
        protected bool __onDisplay__;
        protected bool __finishingTouchesApplied__;
        private int __controlCount__;

        #endregion

        #region Accessors

        /// <summary>
        /// Manager
        /// </summary>
        public PluginManager Manager { get { return __manager__; } set { __manager__ = value; } }
        /// <summary>
        /// Parent
        /// </summary>
        public StructNode Parent { get { return __parent__; } set { __parent__ = value; } }
        /// <summary>
        /// NodeType
        /// </summary>
        public NodeType NodeType { get { return __nodeType__; } set { __nodeType__ = value; } }
        /// <summary>
        /// Name
        /// </summary>
        public string Name { get { return __name__; } set { __name__ = value; } }
        /// <summary>
        /// Offset
        /// </summary>
        public int Offset { get { return __offset__; } set { __offset__ = value; } }
        /// <summary>
        /// Size
        /// </summary>
        public int Size { get { return __size__; } set { __size__ = value; } }
        /// <summary>
        /// Visible
        /// </summary>
        public bool Visible { get { return __visible__; } set { __visible__ = value; } }
        /// <summary>
        /// OnDisplay
        /// </summary>
        protected bool OnDisplay { get { return __onDisplay__; } set { __onDisplay__ = value; } }
        /// <summary>
        /// FinishingTouchesApplied
        /// </summary>
        protected bool FinishingTouchesApplied { get { return __finishingTouchesApplied__; } set { __finishingTouchesApplied__ = value; } }
        /// <summary>
        /// ControlCount
        /// </summary>
        public int ControlCount { get { return __controlCount__; } set { __controlCount__ = value; } }

        #endregion

        public bool IsVisible()
        {
            bool isVis = PluginManager.IsVisible(NodeType);
            if (Visible) return isVis;
            else
            {
                if (PluginManager.IsVisible(NodeType.VisibleFalse)) return isVis;
                else return false;
            }
        }

        public virtual void Read(BitReader br, int seekOffset) { }
        public virtual void Write(BitWriter bw, int seekOffset) { }
        public virtual void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos) { }
    }

    public class StructNode : BaseNode
    {
        #region Instance Data

        private Color __clr__;
        private bool __enabled__;
        private int __chunkCount__;
        private int __translation__;
        private int __currentChunk__;
        private Panel __panel1__;
        private Label __label1__;
        private ComboBox __comboBox1__;

        #endregion

        //private const string labelText = "{0} - Offset: {1} - MapT: {2} - MetaT: {3}";
        private const string labelText = "{0}";

        #region Accessors

        /// <summary>
        /// Clr
        /// </summary>
        public Color Clr { get { return __clr__; } set { __clr__ = value; } }
        /// <summary>
        /// Enabled
        /// </summary>
        public bool Enabled { get { return __enabled__; } set { __enabled__ = value; } }
        /// <summary>
        /// ChunkCount
        /// </summary>
        public int ChunkCount { get { return __chunkCount__; } set { __chunkCount__ = value; } }
        /// <summary>
        /// Translation
        /// </summary>
        public int Translation { get { return __translation__; } set { __translation__ = value; } }
        /// <summary>
        /// CurrentChunk
        /// </summary>
        public int CurrentChunk { get { return __currentChunk__; } set { __currentChunk__ = value; } }
        /// <summary>
        /// Panel1
        /// </summary>
        public Panel Panel1 { get { return __panel1__; } set { __panel1__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// ComboBox1
        /// </summary>
        public ComboBox ComboBox1 { get { return __comboBox1__; } set { __comboBox1__ = value; } }

        #endregion

        public StructNode(PluginManager Manager, StructNode Parent, string Name, int Offset, int Size, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.Reflexive;

            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = Size;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 1;
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            ChunkCount = br.ReadInt32();
            Translation = br.ReadInt32() - Manager.Map.Header.MapMagic;

            CurrentChunk = Translation;
            Enabled = (ChunkCount > 0 && Translation > 0 && Translation <= Manager.Map.Header.Filesize);

            if (OnDisplay) RefreshDisplay();
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            // BG Colour
            Color baseClr = (Parent == null ? PluginManager.ReflexiveBaseColour : Parent.Clr);
            int tmpR = baseClr.R + PluginManager.ReflexiveRedShift;
            if (tmpR > 255) tmpR = PluginManager.ReflexiveBaseColour.R;
            int tmpG = baseClr.G + PluginManager.ReflexiveGreenShift;
            if (tmpG > 255) tmpG = PluginManager.ReflexiveBaseColour.G;
            int tmpB = baseClr.B + PluginManager.ReflexiveBlueShift;
            if (tmpB > 255) tmpB = PluginManager.ReflexiveBaseColour.B;
            Clr = Color.FromArgb(tmpR, tmpG, tmpB);

            // Panel.
            if (Panel1 == null) Panel1 = new Panel();
            else Panel1.Controls.Clear();
            Panel1.Left = xpos;
            Panel1.Top = ypos;
            Panel1.BackColor = Clr;
            Panel1.BorderStyle = BorderStyle.FixedSingle; // Fixed3D;
            collection[cIndex++] = Panel1;

            int ypos2 = 6;

            // Label.
            if (Label1 == null) Label1 = new Label();
            //Label1.Font = new Font(Label1.Font.FontFamily, 9);
            //Label1.Text = string.Format(labelText, this.Name, this.Offset, this.Translation, this.Translation - Manager.Map.Tags.SelectedTag.MetaOffset);
            Label1.Text = this.Name;
            Label1.Left = xpos;
            Label1.Top = ypos2;
            Label1.Size = Label1.PreferredSize;
            Panel1.Controls.Add(Label1);

            ypos2 += Label1.Height + 3;

            // ComboBox.
            if (ComboBox1 == null) ComboBox1 = new ComboBox();
            ComboBox1.Tag = 0;
            ComboBox1.Left = xpos;
            ComboBox1.Top = ypos2;
            ComboBox1.Width = Panel1.Width - (xpos * 2);
            ComboBox1.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
            Panel1.Controls.Add(ComboBox1);

            ypos2 += ComboBox1.Height + 12;
            Panel1.Height = ypos2;

            ypos += Panel1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            ComboBox1.DropDownStyle = ComboBoxStyle.DropDownList;
            ComboBox1.GotFocus += new EventHandler(ComboBox1_GotFocus);
            ComboBox1.SelectedIndexChanged += new EventHandler(ComboBox1_SelectedIndexChanged);

            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            Panel1.Enabled = Enabled;
            if (Enabled)
            {
                if (ComboBox1.Tag != null && ComboBox1.Items.Count != ChunkCount)
                {
                    ComboBox1.Items.Clear();
                    ComboBox1.Items.Add("0");
                    ComboBox1.SelectedIndex = 0;
                    ComboBox1.Tag = null;
                }
                else
                {
                    ComboBox1_SelectedIndexChanged(null, null);
                }
            }
            else
            {
                Panel1.Enabled = false;
                Label1.Text = this.Name;
            }
        }

        private void ComboBox1_GotFocus(object sender, EventArgs e)
        {
            if (ComboBox1.Tag == null)
            {
                object[] objs = new object[ChunkCount - 1];

                for (int x = 0; x < objs.Length; x++)
                {
                    objs[x] = x + 1;
                }
                ComboBox1.Items.AddRange(objs);
                ComboBox1.Tag = 0;
            }
        }

        private void ComboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            BitReader br = new BitReader(Manager.Map.FileName, Endian.Big);
            CurrentChunk = Translation + (ComboBox1.SelectedIndex * this.Size);

            //Label1.Text = string.Format(labelText,
            //        this.Name, this.Offset, this.CurrentChunk, this.CurrentChunk -
            //        Manager.Map.Tags.SelectedTag.MetaOffset);
            Label1.Text = this.Name;

            for (int x = 0; x < Manager.CurrentPlugin.Nodes.Count; x++)
            {
                if (Manager.CurrentPlugin.Nodes[x].Parent == this && Manager.CurrentPlugin.Nodes[x].IsVisible())
                    Manager.CurrentPlugin.Nodes[x].Read(br, CurrentChunk);
            }

            br.Close();
        }
    }

    public class TagRefNode : BaseNode
    {
        #region Instance Data

        #region ReadOnly

        /// <summary>
        /// [Readonly] HasTagType
        /// </summary>
        public readonly bool HasTagType;

        #endregion

        private string __tagType__;
        private int __identity__;
        private string __tagPath__;
        private Label __label1__;
        private ComboBox __comboBox1__;
        private ComboBox __comboBox2__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// TagType
        /// </summary>
        public string TagType { get { return __tagType__; } set { __tagType__ = value; } }
        /// <summary>
        /// Identity
        /// </summary>
        public int Identity { get { return __identity__; } set { __identity__ = value; } }
        /// <summary>
        /// TagPath
        /// </summary>
        public string TagPath { get { return __tagPath__; } set { __tagPath__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// ComboBox1
        /// </summary>
        public ComboBox ComboBox1 { get { return __comboBox1__; } set { __comboBox1__ = value; } }
        /// <summary>
        /// ComboBox2
        /// </summary>
        public ComboBox ComboBox2 { get { return __comboBox2__; } set { __comboBox2__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public TagRefNode(PluginManager Manager, StructNode Parent, bool HasTagType, string Name, int Offset, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.TagRef;

            this.HasTagType = HasTagType;
            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = (HasTagType ? 16 : 4);
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 4;
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            if (HasTagType) TagType = br.ReadString(4);
            br.BaseStream.Position += 8;
            Identity = br.ReadInt32();

            if (Identity == -1) TagPath = Tags.NullTagPath;
            else
            {
                int indx = Manager.Map.Tags.IndexOf(Identity);
                if (!HasTagType) TagType = (indx == -1 ? Tags.NullTagType : Manager.Map.Tags.TagsList[indx].Type);
                TagPath = (indx == -1 ? Tags.NullTagPath : Manager.Map.Tags.TagsList[indx].Path);
            }

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;

            if (HasTagType) bw.Write(TagType);
            bw.BaseStream.Position += 8;
            bw.Write(Identity);
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            //FinishingTouchesApplied = false;

            // NodeType.
            if (Label1 == null) Label1 = new Label();
            Label1.Text = "Tag Ref"; // NodeType.ToString();
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            // TagTypes.
            if (ComboBox1 == null) ComboBox1 = new ComboBox();
            ComboBox1.Left = xpos;
            ComboBox1.Top = ypos;
            ComboBox1.Width = 90;
            collection[cIndex++] = ComboBox1;

            xpos += ComboBox1.Width + 2;

            // TagPaths.
            if (ComboBox2 == null) ComboBox2 = new ComboBox();
            ComboBox2.Left = xpos;
            ComboBox2.Top = ypos;
            collection[cIndex++] = ComboBox2;

            xpos += ComboBox2.Width + 2;
            ypos += 3;

            // Name.
            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Name;
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += ComboBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            // TagTypes.
            ComboBox1.DropDownStyle = ComboBoxStyle.DropDownList;
            ComboBox1.Items.Add(Tags.NullTagType);
            ComboBox1.Items.AddRange((object[])Manager.Map.Tags.TagClasses.Clone());
            if ((ComboBox1.SelectedIndex = ComboBox1.FindString(TagType)) == -1)
            {
                ComboBox1.SelectedIndex = 0;
            }
            ComboBox1.SelectedIndexChanged += new EventHandler(ComboBox1_SelectedIndexChanged);

            // TagPaths.
            ComboBox2.DropDownStyle = ComboBoxStyle.DropDownList;
            ComboBox2.DropDownWidth = 500;
            ComboBox2.Items.Clear();
            ComboBox2.SelectedIndex = ComboBox2.Items.Add(TagPath);
            ComboBox2.GotFocus += new EventHandler(ComboBox2_GotFocus);
            ComboBox2.SelectedIndexChanged += new EventHandler(ComboBox2_SelectedIndexChanged);

            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            // Wrong type selected.
            if (ComboBox1.SelectedItem.ToString() != TagType)
            {
                string tmp = TagPath;

                // Which will add null, and TagPath will become null.
                if ((ComboBox1.SelectedIndex = ComboBox1.FindString(TagType)) == -1)
                {
                    ComboBox1.SelectedIndex = 0;
                }
                ComboBox2.Items.Clear();
                ComboBox2.SelectedIndex = ComboBox2.Items.Add(tmp);
            }
            else // Right type selected.
            {
                if (ComboBox2.Items.Count == 1) // Not fully loaded.
                {
                    ComboBox2.Items.Clear();
                    ComboBox2.SelectedIndex = ComboBox2.Items.Add(TagPath);
                }
                else ComboBox2.SelectedIndex = ComboBox2.FindString(TagPath);
            }
        }

        private void ComboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (ComboBox1.SelectedItem != null)
            {
                TagType = ComboBox1.SelectedItem.ToString();
            }
            ComboBox2.Items.Clear();
            ComboBox2.Items.Add(Tags.NullTagType);
            ComboBox2.SelectedIndex = 0;
        }

        private void ComboBox2_GotFocus(object sender, EventArgs e)
        {
            if (ComboBox2.Items.Count == 1)
            {
                ComboBox2.Items.Clear();
                ComboBox2.Sorted = true;
                ComboBox2.Items.AddRange(Manager.Map.Tags.ParseForList(TagType));
                ComboBox2.Sorted = false;
                ComboBox2.Items.Insert(0, Tags.NullTagPath);
                ComboBox2.SelectedIndex = ComboBox2.FindString(TagPath);
            }
        }

        private void ComboBox2_SelectedIndexChanged(object sender, EventArgs e)
        {
            TagPath = ComboBox2.SelectedItem.ToString();
            Identity = (ComboBox2.SelectedIndex == 0 ? -1 :
                Manager.Map.Tags.TagsList[Manager.Map.Tags.IndexOf(TagType, TagPath)].Identity);
        }

        public void Parent_Resized(int width)
        {
            ComboBox2.Width = width - ComboBox2.Left - Label2.Width - 12;
            Label2.Left = ComboBox2.Right + 2;
        }
    }

    /// <summary>
    /// STRUCTURE UNKNOWN.. GUESS WORK IS WHAT FOLLOWS!..
    /// </summary>
    public class StringIDNode : BaseNode
    {
        #region Instance Data

        private short __strIndex__;
        private Label __label1__;
        private ComboBox __comboBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// StrIndex
        /// </summary>
        public short StrIndex { get { return __strIndex__; } set { __strIndex__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// ComboBox1
        /// </summary>
        public ComboBox ComboBox1 { get { return __comboBox1__; } set { __comboBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public StringIDNode(PluginManager Manager, StructNode Parent, string Name, int Offset, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.StringID;

            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = 2;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            StrIndex = br.ReadInt16();
            if (StrIndex < 0 || StrIndex >= Manager.Map.Header.StringTableCount) StrIndex = 0;

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;
            bw.Write(StrIndex);
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            Label1.Text = "S-ID";
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (ComboBox1 == null) ComboBox1 = new ComboBox();
            ComboBox1.Left = xpos;
            ComboBox1.Top = ypos;
            ComboBox1.Width = 255;
            collection[cIndex++] = ComboBox1;

            xpos += ComboBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Name;
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += ComboBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            ComboBox1.DropDownStyle = ComboBoxStyle.DropDownList;
            ComboBox1.GotFocus += new EventHandler(ComboBox1_GotFocus);
            ComboBox1.SelectedIndexChanged += new EventHandler(ComboBox1_SelectedIndexChanged);

            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            if (ComboBox1.Items.Count != Manager.Map.Strings.StringsList.Length && ComboBox1.Items.Count <= 1)
            {
                ComboBox1.Items.Clear();
                ComboBox1.Items.Add(Manager.Map.Strings.StringsList[StrIndex]);
                ComboBox1.SelectedIndex = 0;
            }
            else ComboBox1.SelectedIndex = StrIndex;
        }

        private void ComboBox1_GotFocus(object sender, EventArgs e)
        {
            if (ComboBox1.Items.Count != Manager.Map.Strings.StringsList.Length && ComboBox1.Items.Count <= 1)
            {
                ComboBox1.Items.Clear();
                ComboBox1.Items.AddRange(Manager.Map.Strings.StringsList);
                ComboBox1.SelectedIndex = StrIndex;
            }
        }

        private void ComboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            StrIndex = (short)ComboBox1.SelectedIndex;
        }
    }

    public class ValueNode : BaseNode
    {
        #region Instance Data

        private object __value__;
        private Label __label1__;
        private ExTextBox __textBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// Value
        /// </summary>
        public object Value { get { return __value__; } set { __value__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// TextBox1
        /// </summary>
        public ExTextBox TextBox1 { get { return __textBox1__; } set { __textBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public ValueNode(PluginManager Manager, StructNode Parent, NodeType NodeType, string Name, int Offset, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType;

            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            switch (NodeType)
            {
                case NodeType.Byte: this.Size = 1; break;
                case NodeType.Short: this.Size = 2; break;
                case NodeType.UShort: this.Size = 2; break;
                case NodeType.Long: this.Size = 4; break;
                case NodeType.ULong: this.Size = 4; break;
                case NodeType.Quad: this.Size = 8; break;
                case NodeType.UQuad: this.Size = 8; break;
                case NodeType.Float: this.Size = 4; break;
                case NodeType.Double: this.Size = 8; break;
            }
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            switch (NodeType)
            {
                case NodeType.Byte: Value = br.ReadByte(); break;
                case NodeType.Short: Value = br.ReadInt16(); break;
                case NodeType.UShort: Value = br.ReadUInt16(); break;
                case NodeType.Long: Value = br.ReadInt32(); break;
                case NodeType.ULong: Value = br.ReadUInt32(); break;
                case NodeType.Quad: Value = br.ReadInt64(); break;
                case NodeType.UQuad: Value = br.ReadUInt64(); break;
                case NodeType.Float: Value = br.ReadSingle(); break;
                case NodeType.Double: Value = br.ReadDouble(); break;
            }

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;

            switch (NodeType)
            {
                case NodeType.Byte: bw.Write((byte)Value); break;
                case NodeType.Short: bw.Write((short)Value); break;
                case NodeType.UShort: bw.Write((ushort)Value); break;
                case NodeType.Long: bw.Write((int)Value); break;
                case NodeType.ULong: bw.Write((uint)Value); break;
                case NodeType.Quad: bw.Write((long)Value); break;
                case NodeType.UQuad: bw.Write((ulong)Value); break;
                case NodeType.Float: bw.Write((float)Value); break;
                case NodeType.Double: bw.Write((double)Value); break;
            }
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            Label1.Text = NodeType.ToString();
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (TextBox1 == null) TextBox1 = new ExTextBox();
            TextBox1.Left = xpos;
            TextBox1.Top = ypos;
            collection[cIndex++] = TextBox1;

            xpos += TextBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Name;
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += TextBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            TextBox1.TextChanged += new EventHandler(TextBox1_TextChanged);

            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            TextBox1.Tag = Value.ToString();
            TextBox1.Text = Value.ToString();
        }

        private void TextBox1_TextChanged(object sender, EventArgs e)
        {
            switch (NodeType)
            {
                case NodeType.Byte:
                    {
                        byte tmp;
                        if (byte.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.Short:
                    {
                        short tmp;
                        if (short.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.UShort:
                    {
                        ushort tmp;
                        if (ushort.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.Long:
                    {
                        int tmp;
                        if (int.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.ULong:
                    {
                        uint tmp;
                        if (uint.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.Quad:
                    {
                        long tmp;
                        if (long.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.UQuad:
                    {
                        ulong tmp;
                        if (ulong.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.Float:
                    {
                        float tmp;
                        if (float.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case NodeType.Double:
                    {
                        double tmp;
                        if (double.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = tmp;
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
            }
        }
    }

    public struct OptionNode
    {
        public string Name;
        public int Value;
        public int ListIndex;
        public void SetListIndex(int newListIndex) { ListIndex = newListIndex; }

        public OptionNode(string name, int value)
        {
            Name = name;
            Value = value;
            ListIndex = -1;
        }
    }

    public class BitmaskNode : BaseNode
    {
        #region Instance Data

        private object __mask__;
        private bool[] __bits__;
        private List<OptionNode> __optionNodes__;
        private Label __label1__;
        private CheckedListBox __checkedListBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// Mask
        /// </summary>
        public object Mask { get { return __mask__; } set { __mask__ = value; } }
        /// <summary>
        /// Bits
        /// </summary>
        public bool[] Bits { get { return __bits__; } set { __bits__ = value; } }
        /// <summary>
        /// Options
        /// </summary>
        public List<OptionNode> OptionNodes { get { return __optionNodes__; } set { __optionNodes__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// CheckedListBox1
        /// </summary>
        public CheckedListBox CheckedListBox1 { get { return __checkedListBox1__; } set { __checkedListBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public BitmaskNode(PluginManager Manager, StructNode Parent, string Name, int Offset, int Size, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.Bitmask;

            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = Size;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;

            this.Bits = new bool[Size];
            this.OptionNodes = new List<OptionNode>();
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            switch (this.Size)
            {
                case 8: Mask = br.ReadByte(); break;
                case 16: Mask = br.ReadUInt16(); break;
                case 32: Mask = br.ReadUInt32(); break;
            }

            uint bit = 1;
            uint mask = Convert.ToUInt32(Mask);

            for (int x = 0; x < Size; x++)
            {
                Bits[x] = ((bit & mask) == bit);
                bit <<= 1;
            }

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;

            for (int x = 0; x < CheckedListBox1.Items.Count; x++)
            {
                Bits[x] = CheckedListBox1.GetItemChecked(x);
            }

            uint bit = 1;
            uint mask = 0;

            for (int x = 0; x < this.Size; x++)
            {
                if (Bits[x]) mask += bit;
                bit <<= 1;
            }
            Mask = mask;

            switch (this.Size)
            {
                case 8: bw.Write((byte)Mask); break;
                case 16: bw.Write((ushort)Mask); break;
                case 32: bw.Write((uint)Mask); break;
            }
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            Label1.Text = string.Format("{0}{1}", NodeType, Size);
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (CheckedListBox1 == null) CheckedListBox1 = new CheckedListBox();
            CheckedListBox1.Left = xpos;
            CheckedListBox1.Top = ypos;
            CheckedListBox1.Width = 200;
            collection[cIndex++] = CheckedListBox1;

            xpos += CheckedListBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Name;
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += CheckedListBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void RefreshDisplay()
        {
            //if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
            //    AddFinishingTouches();

            // Clear.
            CheckedListBox1.Items.Clear();

            // Add.
            for (int x = 0; x < Size; x++)
            {
                bool exists = false;
                for (int y = 0; y < OptionNodes.Count; y++)
                {
                    if (OptionNodes[y].Value == x)
                    {
                        OptionNodes[y].SetListIndex(x);
                        CheckedListBox1.Items.Add(OptionNodes[y].Name);
                        exists = true;
                        break;
                    }
                }
                if (!exists && PluginManager.IsVisible(NodeType.VisibleFalse))
                {
                    CheckedListBox1.Items.Add(x);
                }
            }

            // Check.
            if (PluginManager.IsVisible(NodeType.VisibleFalse))
            {
                for (int x = 0; x < CheckedListBox1.Items.Count; x++)
                {
                    CheckedListBox1.SetItemChecked(x, Bits[x]);
                }
            }
            else
            {
                for (int x = 0; x < OptionNodes.Count; x++)
                {
                    CheckedListBox1.SetItemChecked(x, Bits[OptionNodes[x].ListIndex]);
                }
            }
        }

        /// <summary>
        /// REDUND
        /// </summary>
        public CheckBox GetCheckBox(int value)
        {
            CheckBox checkBox = new CheckBox();
            checkBox.AutoSize = true;

            bool exists = false;

            for (int x = 0; x < OptionNodes.Count; x++)
            {
                if (OptionNodes[x].Value == value)
                {
                    checkBox.Text = OptionNodes[x].Name;
                    exists = true;
                    break;
                }
            }

            if (!exists && PluginManager.IsVisible(NodeType.VisibleFalse))
            {
                checkBox.Visible = false;
            }

            return checkBox;
        }
    }

    public class EnumNode : BaseNode
    {
        #region Instance Data

        private object __value__;
        private List<OptionNode> __optionNodes__;
        private Label __label1__;
        private ComboBox __comboBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// Value
        /// </summary>
        public object Value { get { return __value__; } set { __value__ = value; } }
        /// <summary>
        /// Options
        /// </summary>
        public List<OptionNode> OptionNodes { get { return __optionNodes__; } set { __optionNodes__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// ComboBox1
        /// </summary>
        public ComboBox ComboBox1 { get { return __comboBox1__; } set { __comboBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public EnumNode(PluginManager Manager, StructNode Parent, string Name, int Offset, int Size, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.Enum;

            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = Size;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;

            this.OptionNodes = new List<OptionNode>();
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            switch (this.Size)
            {
                case 8: Value = br.ReadByte(); break;
                case 16: Value = br.ReadInt16(); break;
                case 32: Value = br.ReadInt32(); break;
            }

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;

            switch (this.Size)
            {
                case 8: bw.Write((byte)Value); break;
                case 16: bw.Write((short)Value); break;
                case 32: bw.Write((int)Value); break;
            }
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            Label1.Text = string.Format("{0}{1}", NodeType, Size);
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (ComboBox1 == null) ComboBox1 = new ComboBox();
            ComboBox1.Left = xpos;
            ComboBox1.Top = ypos;
            collection[cIndex++] = ComboBox1;

            xpos += ComboBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Name;
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += ComboBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            ComboBox1.DropDownStyle = ComboBoxStyle.DropDownList;
            ComboBox1.SelectedIndexChanged += new EventHandler(ComboBox1_SelectedIndexChanged);
            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            // Clear.
            ComboBox1.Items.Clear();

            // Add.
            for (int x = 0; x < Size; x++)
            {
                bool exists = false;
                for (int y = 0; y < OptionNodes.Count; y++)
                {
                    if (OptionNodes[y].Value == x)
                    {
                        OptionNodes[y].SetListIndex(
                            ComboBox1.Items.Add(string.Format("{0:D2}: {1}", x, OptionNodes[y].Name))
                            );

                        exists = true;
                        break;
                    }
                }
                if (!exists && PluginManager.IsVisible(NodeType.VisibleFalse))
                {
                    ComboBox1.Items.Add(x.ToString("D2"));
                }
            }

            // Check.
            int val = Convert.ToInt32(Value);
            if (val >= 0 && val < ComboBox1.Items.Count)
            {
                ComboBox1.SelectedIndex = val;
                ComboBox1.Enabled = true;
            }
            else
            {
                ComboBox1.Enabled = false;
            }
        }

        private void ComboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            Value = ComboBox1.SelectedIndex;
        }
    }

    public class StringNode : BaseNode
    {
        #region Instance Data

        private string __text__;
        private Encoding __encoding__;
        private Label __label1__;
        private ExTextBox __textBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// Text
        /// </summary>
        public string Text { get { return __text__; } set { __text__ = value; } }
        /// <summary>
        /// Encoding
        /// </summary>
        public Encoding Encoding { get { return __encoding__; } set { __encoding__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// TextBox1
        /// </summary>
        public ExTextBox TextBox1 { get { return __textBox1__; } set { __textBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public StringNode(PluginManager Manager, StructNode Parent, byte Encoder, string Name, int Offset, int Size, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.String;

            switch (Encoder)
            {
                case 0: this.Encoding = Encoding.ASCII; break;
                case 1: this.Encoding = Encoding.Unicode; break;
            }
            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = Size;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;
            Text = br.ReadCString(Size, Encoding);

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;
            bw.Write(Encoding.GetBytes(Text), Size);
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            if (this.Encoding == Encoding.ASCII) Label1.Text = string.Format("ASCII{0}", Size);
            else if (this.Encoding == Encoding.Unicode) Label1.Text = string.Format("Unicode{0}", Size);
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (TextBox1 == null) TextBox1 = new ExTextBox();
            TextBox1.Left = xpos;
            TextBox1.Top = ypos;
            collection[cIndex++] = TextBox1;

            xpos += TextBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Name;
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += TextBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            if (this.Encoding == Encoding.ASCII) TextBox1.MaxLength = Size;
            else if (this.Encoding == Encoding.Unicode) TextBox1.MaxLength = Size / 2;

            TextBox1.TextChanged += new EventHandler(TextBox1_TextChanged);

            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            TextBox1.Text = Text;
        }

        private void TextBox1_TextChanged(object sender, EventArgs e)
        {
            Text = TextBox1.Text;
        }
    }

    public class UnknownNode : BaseNode
    {
        #region Instance Data

        private byte[] __value__;
        private Label __label1__;
        private ExTextBox __textBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// Value
        /// </summary>
        public byte[] Value { get { return __value__; } set { __value__ = value; } }
        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// TextBox1
        /// </summary>
        public ExTextBox TextBox1 { get { return __textBox1__; } set { __textBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public UnknownNode(PluginManager Manager, StructNode Parent, int Offset, int Size, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.Unknown;

            this.Offset = Offset;
            this.Size = Size;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;
        }

        public override void Read(BitReader br, int seekOffset)
        {
            br.BaseStream.Position = seekOffset + Offset;

            Value = br.ReadBytes(Size);

            if (OnDisplay) RefreshDisplay();
        }

        public override void Write(BitWriter bw, int seekOffset)
        {
            bw.BaseStream.Position = seekOffset + Offset;

            bw.Write(Value, Size);
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            Label1.Text = NodeType.ToString();
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (TextBox1 == null) TextBox1 = new ExTextBox();
            switch (Size)
            {
                case 4: TextBox1.ReadOnly = false; break;
                case 8: TextBox1.ReadOnly = false; break;
                default: TextBox1.ReadOnly = true; break;
            }
            TextBox1.Left = xpos;
            TextBox1.Top = ypos;
            collection[cIndex++] = TextBox1;

            xpos += TextBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = this.Offset.ToString();
            switch (Size)
            {
                case 4: Label2.Text += " (F)"; break;
                case 8: Label2.Text += " (D)"; break;
                default: Label2.Text += " (BA)"; break;
            }
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += TextBox1.Height + 6;

            OnDisplay = true;
            if (Parent == null || Parent.Enabled) RefreshDisplay();
        }

        public void AddFinishingTouches()
        {
            TextBox1.TextChanged += new EventHandler(TextBox1_TextChanged);

            FinishingTouchesApplied = true;
        }

        public void RefreshDisplay()
        {
            if (!FinishingTouchesApplied && (Parent == null || Parent.Enabled))
                AddFinishingTouches();

            switch (Size)
            {
                case 4:
                    TextBox1.Tag = BitConverter.ToSingle(Value, 0).ToString();
                    TextBox1.Text = TextBox1.Tag.ToString();
                    break;

                case 8:
                    TextBox1.Tag = BitConverter.ToDouble(Value, 0).ToString();
                    TextBox1.Text = TextBox1.Tag.ToString();
                    break;

                default:
                    TextBox1.Tag = BitConverter.ToString(Value);
                    TextBox1.Text = TextBox1.Tag.ToString();
                    break;
            }
        }

        private void TextBox1_TextChanged(object sender, EventArgs e)
        {
            switch (Size)
            {
                case 4:
                    {
                        float tmp;
                        if (float.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = BitConverter.GetBytes(tmp);
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                case 8:
                    {
                        double tmp;
                        if (double.TryParse(TextBox1.Text, out tmp))
                        {
                            Value = BitConverter.GetBytes(tmp);
                            TextBox1.Tag = TextBox1.Text;
                        }
                        else TextBox1.Text = TextBox1.Tag.ToString();
                        break;
                    }
                default:
                    {
                        break;
                    }
            }
        }
    }

    public class VoidNode : BaseNode
    {
        #region Instance Data

        private Label __label1__;
        private ExTextBox __textBox1__;
        private Label __label2__;

        #endregion

        #region Accessors

        /// <summary>
        /// Label1
        /// </summary>
        public Label Label1 { get { return __label1__; } set { __label1__ = value; } }
        /// <summary>
        /// TextBox1
        /// </summary>
        public ExTextBox TextBox1 { get { return __textBox1__; } set { __textBox1__ = value; } }
        /// <summary>
        /// Label2
        /// </summary>
        public Label Label2 { get { return __label2__; } set { __label2__ = value; } }

        #endregion

        public VoidNode(PluginManager Manager, StructNode Parent, string Name, int Offset, int Size, bool Visible)
        {
            this.Manager = Manager;
            this.Parent = Parent;
            this.NodeType = NodeType.Void;

            this.Name = (Name == null ? string.Empty : Name);
            this.Offset = Offset;
            this.Size = Size;
            this.Visible = Visible;

            this.OnDisplay = false;
            this.FinishingTouchesApplied = false;
            this.ControlCount = 3;
        }

        public override void ParseForViewing(Control[] collection, ref int cIndex, int xpos, ref int ypos)
        {
            if (Label1 == null) Label1 = new Label();
            Label1.Text = NodeType.ToString();
            Label1.Left = xpos;
            Label1.Top = ypos;
            Label1.Size = Label1.PreferredSize;
            collection[cIndex++] = Label1;

            xpos += PluginManager.BoxUsersStart;
            ypos -= 3;

            if (TextBox1 == null) TextBox1 = new ExTextBox();
            TextBox1.Enabled = false;
            TextBox1.Left = xpos;
            TextBox1.Top = ypos;
            collection[cIndex++] = TextBox1;

            xpos += TextBox1.Width + 2;
            ypos += 3;

            if (Label2 == null) Label2 = new Label();
            Label2.Text = string.Format("{0}({1}B)", this.Name, this.Size);
            Label2.Left = xpos;
            Label2.Top = ypos;
            Label2.Size = Label2.PreferredSize;
            collection[cIndex++] = Label2;

            ypos += TextBox1.Height + 6;

            OnDisplay = true;
            //if (Parent == null || Parent.Enabled) RefreshDisplay();
        }
    }
}