using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;
using System.Extended;
using System.CodeDom.Compiler;
using H3UtilityToolkit.Misc;

namespace H3UtilityToolkit.PluginStuffs
{
    /// <summary>
    /// Converts between the various plugin formats.
    /// </summary>
    public sealed class PluginConverter
    {
        #region Instance Data

        private static PluginManager.PluginInfo __convertInfo__;

        #endregion

        #region Accessors

        /// <summary>
        /// [static] ConvertInfo
        /// </summary>
        public static PluginManager.PluginInfo ConvertInfo { get { return __convertInfo__; } set { __convertInfo__ = value; } }

        #endregion

        static PluginConverter()
        {
            ConvertInfo = new PluginManager.PluginInfo();
        }

        #region Load

        // Loading done in PluginManager.cs

        #endregion

        #region Save

        /// <summary>
        /// Saves loaded plugin as a .dta at the specified save location. 
        /// </summary>
        /// <param name="fileName">The specified save location. </param>
        public static void SaveAsDefinedThroughAttributesPlugin(string fileName)
        {
            XmlTextWriter xw = new XmlTextWriter(fileName, Encoding.Default);
            xw.Formatting = Formatting.Indented;

            #region Constants

            const string dta_plugin = "plugin";

            const string att_class = "class";
            const string att_headerLength = "headerlen";

            const string dta_revision = "revision";
            const string att_author = "author";
            const string att_version = "version";
            const string att_comment = "comment";

            const string att_name = "name";
            const string att_offset = "offset";
            const string att_size = "size";
            const string att_visible = "visible";

            const string dta_tagRef = "tagref";
            const string att_hastagtype = "hastagtype";

            const string dta_struct = "struct";
            const string dta_stringID = "stringid";
            const string dta_void = "void";
            const string dta_unknown = "unknown";
            const string dta_byte = "byte";
            const string dta_short = "short";
            const string dta_ushort = "ushort";
            const string dta_long = "long";
            const string dta_ulong = "ulong";
            const string dta_quad = "quad";
            const string dta_uquad = "uquad";
            const string dta_float = "float";
            const string dta_double = "double";

            const string dta_option = "option";
            const string att_value = "value";

            const string dta_ascii = "ascii";
            const string dta_unicode = "unicode";

            #endregion

            xw.WriteStartElement(dta_plugin);
            xw.WriteAttributeString(att_class, ConvertInfo.Class);
            xw.WriteAttributeString(att_headerLength, ConvertInfo.HeaderLength.ToString());

            for (int x = 0; x < ConvertInfo.Revisions.Count; x++)
            {
                xw.WriteStartElement(dta_revision);
                xw.WriteAttributeString(att_author, ConvertInfo.Revisions[x].Author);
                xw.WriteAttributeString(att_version, ConvertInfo.Revisions[x].Version);
                xw.WriteAttributeString(att_comment, ConvertInfo.Revisions[x].Comment);
                xw.WriteEndElement();
            }

            StructNode curStruct = null;
            for (int x = 0; x < ConvertInfo.Nodes.Count; x++)
            {
                if (ConvertInfo.Nodes[x].NodeType != NodeType.Reflexive || ConvertInfo.Nodes[x].Parent != curStruct)
                {
                    while (curStruct != ConvertInfo.Nodes[x].Parent)
                    {
                        xw.WriteEndElement();
                        curStruct = curStruct.Parent;
                    }
                }
                switch (ConvertInfo.Nodes[x].NodeType)
                {
                    case NodeType.Reflexive:
                        {
                            xw.WriteStartElement(dta_struct);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_size, ConvertInfo.Nodes[x].Size.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());

                            curStruct = (ConvertInfo.Nodes[x] as StructNode);
                            //xw.WriteEndElement();
                            break;
                        }
                    case NodeType.TagRef:
                        {
                            xw.WriteStartElement(dta_tagRef);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_hastagtype, (ConvertInfo.Nodes[x] as TagRefNode).HasTagType.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.StringID:
                        {
                            xw.WriteStartElement(dta_stringID);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Void:
                        {
                            xw.WriteStartElement(dta_void);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name.ToString());
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_size, ConvertInfo.Nodes[x].Size.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Unknown:
                        {
                            xw.WriteStartElement(dta_unknown);
                            //xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_size, ConvertInfo.Nodes[x].Size.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Byte:
                        {
                            xw.WriteStartElement(dta_byte);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Short:
                        {
                            xw.WriteStartElement(dta_short);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.UShort:
                        {
                            xw.WriteStartElement(dta_ushort);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Long:
                        {
                            xw.WriteStartElement(dta_long);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.ULong:
                        {
                            xw.WriteStartElement(dta_ulong);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Quad:
                        {
                            xw.WriteStartElement(dta_quad);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.UQuad:
                        {
                            xw.WriteStartElement(dta_uquad);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Float:
                        {
                            xw.WriteStartElement(dta_float);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Double:
                        {
                            xw.WriteStartElement(dta_double);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Bitmask:
                        {
                            xw.WriteStartElement(string.Format("bitmask{0}", ConvertInfo.Nodes[x].Size));
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            foreach (OptionNode on in (ConvertInfo.Nodes[x] as BitmaskNode).OptionNodes)
                            {
                                xw.WriteStartElement(dta_option);
                                xw.WriteAttributeString(att_name, on.Name);
                                xw.WriteAttributeString(att_value, on.Value.ToString());
                                xw.WriteEndElement();
                            }
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Enum:
                        {
                            xw.WriteStartElement(string.Format("enum{0}", ConvertInfo.Nodes[x].Size));
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            foreach (OptionNode on in (ConvertInfo.Nodes[x] as EnumNode).OptionNodes)
                            {
                                xw.WriteStartElement(dta_option);
                                xw.WriteAttributeString(att_name, on.Name);
                                xw.WriteAttributeString(att_value, on.Value.ToString());
                                xw.WriteEndElement();
                            }
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.String:
                        {
                            if ((ConvertInfo.Nodes[x] as StringNode).Encoding == Encoding.ASCII)
                            {
                                xw.WriteStartElement(dta_ascii);
                            }
                            else // Unicode.
                            {
                                xw.WriteStartElement(dta_unicode);
                            }

                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_size, ConvertInfo.Nodes[x].Size.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();

                            break;
                        }
                }
            }

            xw.WriteEndElement();
            xw.Close();
        }

        /// <summary>
        /// Saves loaded plugin as a .ent at the specified save location. 
        /// </summary>
        /// <param name="fileName">The specified save location. </param>
        public static void SaveAsEntityPlugin(string fileName)
        {
            XmlTextWriter xw = new XmlTextWriter(fileName, Encoding.Default);
            xw.Formatting = Formatting.Indented;

            #region Constants

            const string ent_plugin = "plugin";

            const string att_class = "class";
            const string att_author = "author";
            const string att_version = "version";
            const string att_headerSize = "headersize";

            const string ent_revision = "revision";

            const string ent_struct = "struct";
            const string ent_tagRef = "tag";
            const string ent_id = "id";
            const string ent_stringID = "stringid";
            const string ent_unused = "unused";
            const string ent_undefined = "undefined";
            const string ent_byte = "byte";
            const string ent_short = "short";
            const string ent_ushort = "ushort";
            const string ent_int = "int";
            const string ent_uint = "uint";
            const string ent_float = "float";

            const string ent_option = "option";
            const string att_value = "value";

            const string att_name = "name";
            const string att_offset = "offset";
            const string att_size = "size";
            const string att_visible = "visible";

            #endregion

            xw.WriteStartElement(ent_plugin);
            xw.WriteAttributeString(att_class, ConvertInfo.Class);
            xw.WriteAttributeString(att_author, ConvertInfo.Author);
            xw.WriteAttributeString(att_version, ConvertInfo.Version);
            xw.WriteAttributeString(att_headerSize, ConvertInfo.HeaderLength.ToString());

            for (int x = 0; x < ConvertInfo.Revisions.Count; x++)
            {
                xw.WriteStartElement(ent_revision);
                xw.WriteAttributeString(att_author, ConvertInfo.Revisions[x].Author);
                xw.WriteAttributeString(att_version, ConvertInfo.Revisions[x].Version);
                xw.WriteString(ConvertInfo.Revisions[x].Comment);
                xw.WriteEndElement();
            }

            StructNode curStruct = null;
            for (int x = 0; x < ConvertInfo.Nodes.Count; x++)
            {
                if (ConvertInfo.Nodes[x].NodeType != NodeType.Reflexive || ConvertInfo.Nodes[x].Parent != curStruct)
                {
                    while (curStruct != ConvertInfo.Nodes[x].Parent)
                    {
                        xw.WriteEndElement();
                        curStruct = curStruct.Parent;
                    }
                }
                switch (ConvertInfo.Nodes[x].NodeType)
                {
                    case NodeType.Reflexive:
                        {
                            //<struct name="Skies11" offset="8" visible="true" size="8">
                            xw.WriteStartElement(ent_struct);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteAttributeString(att_size, ConvertInfo.Nodes[x].Size.ToString());
                            //xw.WriteAttributeString("label",ConvertInfo.PluginNodes[x].Label);
                            /*
                            if (!PluginNodes[x].HasCount)
                            {
                                xw.WriteAttributeString("HasCount",ConvertInfo.PluginNodes[x].HasCount.ToString());
                            }
                            */
                            curStruct = (ConvertInfo.Nodes[x] as StructNode);
                            //xw.WriteEndElement();
                            break;
                        }
                    case NodeType.StringID:
                        {
                            //<stringid name="Unknown" offset="52" visible="false" />
                            xw.WriteStartElement(ent_stringID);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.TagRef:
                        {
                            bool hasTagType = (ConvertInfo.Nodes[x] as TagRefNode).HasTagType;
                            if (hasTagType)
                            {
                                xw.WriteStartElement(ent_tagRef);
                                xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                                xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                                xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                                xw.WriteEndElement();
                            }

                            xw.WriteStartElement(ent_id);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, (ConvertInfo.Nodes[x].Offset + (hasTagType ? 8 : 0)).ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.String:
                        {
                            if ((ConvertInfo.Nodes[x] as StringNode).Encoding == Encoding.ASCII)
                            {
                                xw.WriteStartElement(string.Format("string{0}", ConvertInfo.Nodes[x].Size));
                            }
                            else // Unicode.
                            {
                                xw.WriteStartElement(string.Format("unicode{0}", ConvertInfo.Nodes[x].Size));
                            }

                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();

                            break;
                        }
                    case NodeType.Void:
                        {
                            xw.WriteStartElement(ent_unused);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_size, ConvertInfo.Nodes[x].Size.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Unknown:
                        {
                            xw.WriteStartElement(ent_undefined);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Float:
                        {
                            xw.WriteStartElement(ent_float);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Byte:
                        {
                            xw.WriteStartElement(ent_byte);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            /*
                            if (!PluginNodes[x].entIndex.nulled)
                            {
                                xw.WriteAttributeString("reflexiveoffset",ConvertInfo.PluginNodes[x].entIndex.ReflexiveOffset.ToString());
                                xw.WriteAttributeString("reflexivesize",ConvertInfo.PluginNodes[x].entIndex.ReflexiveSize.ToString());
                                xw.WriteAttributeString("itemoffset",ConvertInfo.PluginNodes[x].entIndex.ItemOffset.ToString());
                                xw.WriteAttributeString("itemtype",ConvertInfo.PluginNodes[x].entIndex.ItemType.ToString());
                                xw.WriteAttributeString("layer",ConvertInfo.PluginNodes[x].entIndex.reflexiveLayer.ToString());
                            }
                            */
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Short:
                        {
                            xw.WriteStartElement(ent_short);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            /*
                            if (!PluginNodes[x].entIndex.nulled)
                            {
                                xw.WriteAttributeString("reflexiveoffset",ConvertInfo.PluginNodes[x].entIndex.ReflexiveOffset.ToString());
                                xw.WriteAttributeString("reflexivesize",ConvertInfo.PluginNodes[x].entIndex.ReflexiveSize.ToString());
                                xw.WriteAttributeString("itemoffset",ConvertInfo.PluginNodes[x].entIndex.ItemOffset.ToString());
                                xw.WriteAttributeString("itemtype",ConvertInfo.PluginNodes[x].entIndex.ItemType.ToString());
                                xw.WriteAttributeString("layer",ConvertInfo.PluginNodes[x].entIndex.reflexiveLayer.ToString());
                            }
                            */
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.UShort:
                        {
                            xw.WriteStartElement(ent_ushort);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            /*
                            if (!PluginNodes[x].entIndex.nulled)
                            {
                                xw.WriteAttributeString("reflexiveoffset",ConvertInfo.PluginNodes[x].entIndex.ReflexiveOffset.ToString());
                                xw.WriteAttributeString("reflexivesize",ConvertInfo.PluginNodes[x].entIndex.ReflexiveSize.ToString());
                                xw.WriteAttributeString("itemoffset",ConvertInfo.PluginNodes[x].entIndex.ItemOffset.ToString());
                                xw.WriteAttributeString("itemtype",ConvertInfo.PluginNodes[x].entIndex.ItemType.ToString());
                                xw.WriteAttributeString("layer",ConvertInfo.PluginNodes[x].entIndex.reflexiveLayer.ToString());
                            }
                            */
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Long:
                        {
                            xw.WriteStartElement(ent_int);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            /*
                            if (!PluginNodes[x].entIndex.nulled)
                            {
                                xw.WriteAttributeString("reflexiveoffset",ConvertInfo.PluginNodes[x].entIndex.ReflexiveOffset.ToString());
                                xw.WriteAttributeString("reflexivesize",ConvertInfo.PluginNodes[x].entIndex.ReflexiveSize.ToString());
                                xw.WriteAttributeString("itemoffset",ConvertInfo.PluginNodes[x].entIndex.ItemOffset.ToString());
                                xw.WriteAttributeString("itemtype",ConvertInfo.PluginNodes[x].entIndex.ItemType.ToString());
                                xw.WriteAttributeString("layer",ConvertInfo.PluginNodes[x].entIndex.reflexiveLayer.ToString());
                            }
                            */
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.ULong:
                        {
                            xw.WriteStartElement(ent_uint);
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            /*
                            if (!PluginNodes[x].entIndex.nulled)
                            {
                                xw.WriteAttributeString("reflexiveoffset",ConvertInfo.PluginNodes[x].entIndex.ReflexiveOffset.ToString());
                                xw.WriteAttributeString("reflexivesize",ConvertInfo.PluginNodes[x].entIndex.ReflexiveSize.ToString());
                                xw.WriteAttributeString("itemoffset",ConvertInfo.PluginNodes[x].entIndex.ItemOffset.ToString());
                                xw.WriteAttributeString("itemtype",ConvertInfo.PluginNodes[x].entIndex.ItemType.ToString());
                                xw.WriteAttributeString("layer",ConvertInfo.PluginNodes[x].entIndex.reflexiveLayer.ToString());
                            }
                            */
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Bitmask:
                        {
                            xw.WriteStartElement(string.Format("bitmask{0}", ConvertInfo.Nodes[x].Size));
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            foreach (OptionNode on in (ConvertInfo.Nodes[x] as BitmaskNode).OptionNodes)
                            {
                                xw.WriteStartElement(ent_option);
                                xw.WriteAttributeString(att_name, on.Name);
                                xw.WriteAttributeString(att_value, on.Value.ToString());
                                xw.WriteEndElement();
                            }
                            xw.WriteEndElement();
                            break;
                        }
                    case NodeType.Enum:
                        {
                            xw.WriteStartElement(string.Format("enum{0}", ConvertInfo.Nodes[x].Size));
                            xw.WriteAttributeString(att_name, ConvertInfo.Nodes[x].Name);
                            xw.WriteAttributeString(att_offset, ConvertInfo.Nodes[x].Offset.ToString());
                            xw.WriteAttributeString(att_visible, ConvertInfo.Nodes[x].Visible.ToString());
                            foreach (OptionNode on in (ConvertInfo.Nodes[x] as EnumNode).OptionNodes)
                            {
                                xw.WriteStartElement(ent_option);
                                xw.WriteAttributeString(att_name, on.Name);
                                xw.WriteAttributeString(att_value, on.Value.ToString());
                                xw.WriteEndElement();
                            }
                            xw.WriteEndElement();
                            break;
                        }
                }
            }

            xw.WriteEndElement();
            xw.Close();
        }

        /// <summary>
        /// Saves loaded plugin as a .cs at the specified save location. 
        /// </summary>
        /// <param name="fileName">The specified save location. </param>
        public static void SaveAsCSharpPlugin(string fileName)
        {
            IndentedTextWriter iw = new IndentedTextWriter(new StreamWriter(fileName), " ");
            iw.Indent = 0;

            // Start.
            iw.WriteLine("using System;");
            iw.WriteLine("using System.Collections.Generic;");
            iw.WriteLine("using System.Text;");
            iw.WriteLine("using Halo3Map.IO;");
            iw.WriteLine("using Halo3Map.Data;");
            iw.WriteLine();
            iw.WriteLine("namespace Halo3Map.Tags.Layouts");
            iw.WriteLine("{");

            iw.Indent = 4;
            iw.WriteLine("[Tag(\"{0}\", \"{1}\", {2})]",
                ConvertInfo.Author,
                MiscUtilities.RemoveIllegalFileNameChars(ConvertInfo.Class),
                double.Parse(ConvertInfo.Version));

            iw.WriteLine("public class {0} : H3Tag", MiscUtilities.RemoveIllegalFileNameChars(ConvertInfo.Class));
            iw.WriteLine("{");

            iw.Indent = 8;

            for (int x = 0; x < ConvertInfo.Revisions.Count; x++)
            {
                iw.WriteLine("// Revision: {0} - {1}", ConvertInfo.Revisions[x].Author, ConvertInfo.Revisions[x].Comment);
            }
            iw.WriteLine();

            StructNode curStruct = null;
            bool newlineEntered = false;

            for (int x = 0; x < ConvertInfo.Nodes.Count; x++)
            {
                if (ConvertInfo.Nodes[x].NodeType != NodeType.Reflexive || ConvertInfo.Nodes[x].Parent != curStruct)
                {
                    int indent = iw.Indent;
                    while (curStruct != ConvertInfo.Nodes[x].Parent)
                    {
                        iw.Indent -= 4;
                        iw.WriteLine("}");
                        curStruct = curStruct.Parent;
                    }
                    if (indent != iw.Indent)
                    {
                        iw.WriteLine();
                        newlineEntered = true;
                    }
                }
                switch (ConvertInfo.Nodes[x].NodeType)
                {
                    case NodeType.Reflexive:
                        {
                            string name_l = MiscUtilities.LowerCaseLetter(0, ConvertInfo.Nodes[x].Name);
                            string name_h = MiscUtilities.CapitaliseLetter(0, ConvertInfo.Nodes[x].Name);

                            if (!newlineEntered) iw.WriteLine();
                            iw.WriteLine("public H3Reflexive<{0}> {1};", name_l, name_h);
                            iw.WriteLine("[Size({0})]", ConvertInfo.Nodes[x].Size);
                            iw.WriteLine("public class {0} : Chunk", name_l);
                            iw.WriteLine("{");

                            iw.Indent += 4;
                            curStruct = (ConvertInfo.Nodes[x] as StructNode);
                            break;
                        }
                    case NodeType.TagRef:
                        {
                            iw.WriteLine("public TagReference {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.StringID:
                        {
                            iw.WriteLine("public StringID {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Void:
                        {
                            iw.WriteLine("[Size({0})]", ConvertInfo.Nodes[x].Size);
                            iw.WriteLine("public byte[] {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Unknown:
                        {
                            iw.WriteLine("[Size({0})]", ConvertInfo.Nodes[x].Size);
                            iw.WriteLine("public byte[] Unknown{0};", ConvertInfo.Nodes[x].Offset);
                            break;
                        }
                    case NodeType.Byte:
                        {
                            iw.WriteLine("public byte {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Short:
                        {
                            iw.WriteLine("public short {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.UShort:
                        {
                            iw.WriteLine("public ushort {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Long:
                        {
                            iw.WriteLine("public int {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.ULong:
                        {
                            iw.WriteLine("public uint {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Quad:
                        {
                            iw.WriteLine("public long {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.UQuad:
                        {
                            iw.WriteLine("public ulong {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Float:
                        {
                            iw.WriteLine("public float {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Double:
                        {
                            iw.WriteLine("public double {0};", ConvertInfo.Nodes[x].Name);
                            break;
                        }
                    case NodeType.Bitmask:
                        {
                            // NOT SUPPORTED
                            break;
                        }
                    case NodeType.Enum:
                        {
                            EnumNode en = (ConvertInfo.Nodes[x] as EnumNode);
                            switch (en.Size)
                            {
                                case 8: iw.WriteLine("public enum {0} : byte", en.Name); break;
                                case 16: iw.WriteLine("public enum {0} : short", en.Name); break;
                                case 32: iw.WriteLine("public enum {0} : int", en.Name); break;
                                case 64: iw.WriteLine("public enum {0} : long", en.Name); break;
                                default: goto End;
                            }
                            iw.WriteLine("{");
                            if (en.OptionNodes.Count != 0)
                            {
                                iw.Indent += 4;
                                int indx;
                                for (indx = 0; indx < en.OptionNodes.Count - 1; indx++)
                                {
                                    iw.WriteLine("{0} = {1},", en.OptionNodes[indx].Name, en.OptionNodes[indx].Value);
                                }
                                iw.WriteLine("{0} = {1}", en.OptionNodes[indx].Name, en.OptionNodes[indx].Value);
                                iw.Indent -= 4;
                            }
                            iw.WriteLine("}");

                        End:
                            break;
                        }
                    case NodeType.String:
                        {
                            StringNode sn = (ConvertInfo.Nodes[x] as StringNode);
                            if (sn.Encoding == Encoding.ASCII) iw.WriteLine("[String(StringType.Ascii, {0})]", sn.Size);
                            else if (sn.Encoding == Encoding.Unicode) iw.WriteLine("[String(StringType.Unicode, {0})]", sn.Size);
                            else goto End;

                            iw.WriteLine("public string {0};", sn.Name);

                        End:
                            break;
                        }
                }

                newlineEntered = false;
            }

            while (iw.Indent != 0)
            {
                iw.Indent -= 4;
                iw.WriteLine("}");
            }
            iw.Close();
        }

        #endregion
    }
}