using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Drawing;
using Microsoft.CSharp;
using System.CodeDom.Compiler;
using System.Reflection;
using System.IO;
using System.Extended;
using Halo3Map.Data;
using H3UtilityToolkit;
using H3UtilityToolkit.Misc;
using H3UtilityToolkit.Essentials;

namespace H3UtilityToolkit.PluginStuffs
{
    /// <summary>
    /// Manages plugin I/O operations.
    /// </summary>
    public class PluginManager
    {
        #region Instance Data

        #region Constants

        /// <summary>
        /// [Const] DtaExtension
        /// </summary>
        public const string DtaExtension = ".dta";
        /// <summary>
        /// [Const] EntExtension
        /// </summary>
        public const string EntExtension = ".ent";
        /// <summary>
        /// [Const] CSExtension
        /// </summary>
        public const string CSExtension = ".cs";

        #endregion

        private Map __map__;
        private PluginInfo __currentPlugin__;
        private static string __pluginDirectory__;
        private static NodeType __visibleMask__;
        private static int __boxUsersStart__;
        private static Color __reflexiveBaseColour__;
        private static int __reflexiveRedShift__;
        private static int __reflexiveGreenShift__;
        private static int __reflexiveBlueShift__;
        private static PluginConverter __pluginConverter__;

        #endregion

        #region Accessors

        /// <summary>
        /// Map
        /// </summary>
        public Map Map { get { return __map__; } set { __map__ = value; } }
        /// <summary>
        /// CurrentPlugin
        /// </summary>
        public PluginInfo CurrentPlugin { get { return __currentPlugin__; } set { __currentPlugin__ = value; } }
        /// <summary>
        /// [static] PluginDirectory
        /// </summary>
        public static string PluginDirectory { get { return __pluginDirectory__; } set { __pluginDirectory__ = value; } }
        /// <summary>
        /// [static] VisibleMask
        /// </summary>
        public static NodeType VisibleMask { get { return __visibleMask__; } set { __visibleMask__ = value; } }
        /// <summary>
        /// [static] BoxUsersStart
        /// </summary>
        public static int BoxUsersStart { get { return __boxUsersStart__; } set { __boxUsersStart__ = value; } }
        /// <summary>
        /// [static] ReflexiveBaseBackGround
        /// </summary>
        public static Color ReflexiveBaseColour { get { return __reflexiveBaseColour__; } set { __reflexiveBaseColour__ = value; } }
        /// <summary>
        /// [static] ReflexiveRedShift
        /// </summary>
        public static int ReflexiveRedShift { get { return __reflexiveRedShift__; } set { __reflexiveRedShift__ = value; } }
        /// <summary>
        /// [static] ReflexiveGreenShift
        /// </summary>
        public static int ReflexiveGreenShift { get { return __reflexiveGreenShift__; } set { __reflexiveGreenShift__ = value; } }
        /// <summary>
        /// [static] ReflexiveBlueShift
        /// </summary>
        public static int ReflexiveBlueShift { get { return __reflexiveBlueShift__; } set { __reflexiveBlueShift__ = value; } }
        /// <summary>
        /// PluginConverter
        /// </summary>
        private static PluginConverter PluginConverter { get { return __pluginConverter__; } set { __pluginConverter__ = value; } }

        #endregion

        static PluginManager()
        {
            PluginDirectory = Application.StartupPath + @"\Plugins\";
#if DEBUG
            PluginDirectory = Path.GetDirectoryName(Application.StartupPath) +
              @"\Release\Plugins\";
#endif
            PluginConverter = new PluginConverter();
            GetBoxUsersStart();
        }

        /// <summary>
        /// Initialises a new instance of the PluginManager class based on the
        /// supplied Map reference.
        /// </summary>
        /// <param name="map"></param>
        public PluginManager(Map map)
        {
            this.Map = map;
            CurrentPlugin = new PluginInfo();
        }

        #region Read

        /// <summary>
        /// Finds and reads the plugin associated with the specified type.
        /// </summary>
        /// <param name="type">The specified type. </param>
        public bool ReadPlugin(string type)
        {
            string fileName = string.Format("{0}{1}", PluginDirectory,
                MiscUtilities.RemoveIllegalFileNameChars(type));

            CurrentPlugin.Clear();

            try
            {
                // Dta
                CurrentPlugin.FileName = fileName + DtaExtension;
                if (File.Exists(CurrentPlugin.FileName))
                {
                    return ReadDefinedThroughAttributesPlugin(this, CurrentPlugin);
                }

                // Ent
                CurrentPlugin.FileName = fileName + EntExtension;
                if (File.Exists(CurrentPlugin.FileName))
                {
                    return ReadEntityPlugin(this, CurrentPlugin);
                }

                // CS
                CurrentPlugin.FileName = fileName + CSExtension;
                if (File.Exists(CurrentPlugin.FileName))
                {
                    return ReadCSharpPlugin(this, CurrentPlugin);
                }
            }
            catch { }

            // else..
            CurrentPlugin.FileName = string.Empty;
            return false;
        }

        private static bool ReadDefinedThroughAttributesPlugin(PluginManager man, PluginInfo pInfo)
        {
            XmlTextReader xr = new XmlTextReader(new FileStream(pInfo.FileName,
                FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            xr.WhitespaceHandling = WhitespaceHandling.None;

            StructNode curParent = null;
            BitmaskNode curBitmask = null;
            EnumNode curEnum = null;

            #region Constants

            const string dta_plugin = "plugin";

            const string att_class = "class";
            const string att_headerLen = "headerlen";

            const string dta_revision = "revision";
            const string att_author = "author";
            const string att_version = "version";
            const string att_comment = "comment";

            const string att_name = "name";
            const string att_offset = "offset";
            const string att_size = "size";
            const string att_visible = "visible";

            const string dta_tagRef = "tagref";
            const string att_hastagtype = "hastagtype";

            const string dta_struct = "struct";
            const string dta_stringID = "stringid";
            const string dta_void = "void";
            const string dta_unknown = "unknown";
            const string dta_byte = "byte";
            const string dta_short = "short";
            const string dta_ushort = "ushort";
            const string dta_long = "long";
            const string dta_ulong = "ulong";
            const string dta_quad = "quad";
            const string dta_uquad = "uquad";
            const string dta_float = "float";
            const string dta_double = "double";

            const string dta_option = "option";
            const string att_value = "value";

            const string dta_bitmask8 = "bitmask8";
            const string dta_bitmask16 = "bitmask16";
            const string dta_bitmask32 = "bitmask32";
            const string dta_enum8 = "enum8";
            const string dta_enum16 = "enum16";
            const string dta_enum32 = "enum32";
            const string dta_ascii = "ascii";
            const string dta_unicode = "unicode";

            #endregion

            // Loops the entire dta file.
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            string xrNameLower = xr.Name.ToLower();
                            switch (xrNameLower)
                            {
                                case dta_plugin:
                                    {
                                        pInfo.Class = xr.GetAttribute(att_class);
                                        pInfo.Author = string.Empty;
                                        pInfo.HeaderLength = ToInt32(xr.GetAttribute(att_headerLen));
                                        break;
                                    }
                                case dta_revision:
                                    {
                                        pInfo.Revisions.Add(new Revision(
                                            xr.GetAttribute(att_author),
                                            xr.GetAttribute(att_version),
                                            xr.GetAttribute(att_comment)));
                                        break;
                                    }
                                case dta_struct:
                                    {
                                        pInfo.Nodes.Add(curParent = new StructNode(
                                            man,
                                            curParent,
                                            xr.GetAttribute(att_name),
                                            ToInt32(xr.GetAttribute(att_offset)),
                                            ToInt32(xr.GetAttribute(att_size)),
                                            ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_tagRef:
                                    {
                                        pInfo.Nodes.Add(new TagRefNode(
                                            man,
                                            curParent,
                                            ToBool(xr.GetAttribute(att_hastagtype)),
                                            xr.GetAttribute(att_name),
                                            ToInt32(xr.GetAttribute(att_offset)),
                                            ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_stringID:
                                    {
                                        pInfo.Nodes.Add(new StringIDNode(
                                            man,
                                            curParent,
                                            xr.GetAttribute(att_name),
                                            ToInt32(xr.GetAttribute(att_offset)),
                                            ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_void:
                                    {
                                        pInfo.Nodes.Add(new VoidNode(
                                            man,
                                            curParent,
                                            xr.GetAttribute(att_name),
                                            ToInt32(xr.GetAttribute(att_offset)),
                                            ToInt32(xr.GetAttribute(att_size)),
                                            ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_unknown:
                                    {
                                        pInfo.Nodes.Add(new UnknownNode(
                                            man,
                                            curParent,
                                            ToInt32(xr.GetAttribute(att_offset)),
                                            ToInt32(xr.GetAttribute(att_size)),
                                            ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_byte:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Byte,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_short:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                              man,
                                              curParent,
                                              NodeType.Short,
                                              xr.GetAttribute(att_name),
                                              ToInt32(xr.GetAttribute(att_offset)),
                                              ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_ushort:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.UShort,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_long:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Long,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_ulong:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.ULong,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_quad:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Quad,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_uquad:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.UQuad,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_float:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Float,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_double:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Double,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_option:
                                    {
                                        OptionNode on = new OptionNode(
                                            xr.GetAttribute(att_name),
                                            ToInt32(xr.GetAttribute(att_value)));

                                        if (curBitmask != null) curBitmask.OptionNodes.Add(on);
                                        else if (curEnum != null) curEnum.OptionNodes.Add(on);

                                        break;
                                    }
                                case dta_bitmask8:
                                    {
                                        pInfo.Nodes.Add(curBitmask = new BitmaskNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             8,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_bitmask16:
                                    {
                                        pInfo.Nodes.Add(curBitmask = new BitmaskNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             16,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_bitmask32:
                                    {
                                        pInfo.Nodes.Add(curBitmask = new BitmaskNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             32,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_enum8:
                                    {
                                        pInfo.Nodes.Add(curEnum = new EnumNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             8,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_enum16:
                                    {
                                        pInfo.Nodes.Add(curEnum = new EnumNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             16,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_enum32:
                                    {
                                        pInfo.Nodes.Add(curEnum = new EnumNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             32,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_ascii:
                                    {
                                        pInfo.Nodes.Add(new StringNode(
                                             man,
                                             curParent,
                                             0,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToInt32(xr.GetAttribute(att_size)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case dta_unicode:
                                    {
                                        pInfo.Nodes.Add(new StringNode(
                                             man,
                                             curParent,
                                             1,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToInt32(xr.GetAttribute(att_size)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                            }
                            break;
                        }
                    case XmlNodeType.EndElement:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case dta_bitmask8:
                                case dta_bitmask16:
                                case dta_bitmask32:
                                    {
                                        curBitmask = null;
                                        break;
                                    }

                                case dta_enum8:
                                case dta_enum16:
                                case dta_enum32:
                                    {
                                        curEnum = null;
                                        break;
                                    }

                                case dta_struct:
                                    {
                                        curParent = curParent.Parent;
                                        break;
                                    }
                            }
                            break;
                        }
                }
            }

            if (pInfo.Revisions.Count != 0)
            {
                pInfo.Version = pInfo.Revisions[0].Version;
            }
            xr.Close();
            return true;
        }

        private static bool ReadEntityPlugin(PluginManager man, PluginInfo pInfo)
        {
            XmlTextReader xr = new XmlTextReader(new FileStream(pInfo.FileName,
                FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite));

            xr.WhitespaceHandling = WhitespaceHandling.None;

            StructNode curParent = null;
            BitmaskNode curBitmask = null;
            EnumNode curEnum = null;
            bool tagJustRead = false;
            bool unusedJustRead = false;

            #region Constants

            const string ent_plugin = "plugin";

            const string att_class = "class";
            const string att_author = "author";
            const string att_version = "version";
            const string att_headerSize = "headersize";

            const string ent_revision = "revision";

            const string ent_struct = "struct";
            const string ent_tagRef = "tag";
            const string ent_id = "id";
            const string ent_script = "script";
            const string ent_stringID = "stringid";
            const string ent_unused = "unused";
            const string ent_undefined = "undefined";
            const string ent_byte = "byte";
            const string ent_short = "short";
            const string ent_ushort = "ushort";
            const string ent_int = "int";
            const string ent_uint = "uint";
            const string ent_float = "float";
            const string ent_option = "option";
            const string att_value = "value";
            const string ent_bitmask8 = "bitmask8";
            const string ent_bitmask16 = "bitmask16";
            const string ent_bitmask = "bitmask";
            const string ent_bitmask32 = "bitmask32";
            const string ent_enum8 = "enum8";
            const string ent_enum16 = "enum16";
            const string ent_enum = "enum";
            const string ent_enum32 = "enum32";
            const string ent_string32 = "string32";
            const string ent_string256 = "string256";
            const string ent_unicode64 = "unicode64";
            const string ent_unicode256 = "unicode256";

            const string att_name = "name";
            const string att_offset = "offset";
            const string att_size = "size";
            const string att_visible = "visible";

            #endregion

            // Loops the entire ent file.
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            string xrNameLower = xr.Name.ToLower();
                            switch (xrNameLower)
                            {
                                case ent_plugin:
                                    {
                                        pInfo.Class = xr.GetAttribute(att_class);
                                        pInfo.Author = xr.GetAttribute(att_author);
                                        pInfo.Version = xr.GetAttribute(att_version);
                                        pInfo.HeaderLength = ToInt32(xr.GetAttribute(att_headerSize));
                                        break;
                                    }
                                case ent_revision:
                                    {
                                        pInfo.Revisions.Add(new Revision(
                                            xr.GetAttribute(att_author),
                                            xr.GetAttribute(att_version),
                                            xr.ReadString()));
                                        break;
                                    }
                                case ent_struct:
                                    {
                                        pInfo.Nodes.Add(curParent = new StructNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToInt32(xr.GetAttribute(att_size)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_tagRef:
                                    {
                                        tagJustRead = true;
                                        pInfo.Nodes.Add(new TagRefNode(
                                             man,
                                             curParent,
                                             true,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_id:
                                    {
                                        if (tagJustRead) break;
                                        if (unusedJustRead)
                                        {
                                            pInfo.Nodes.RemoveAt(pInfo.Nodes.Count - 1);
                                            break;
                                        }

                                        pInfo.Nodes.Add(new TagRefNode(
                                             man,
                                             curParent,
                                             false,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_script:
                                case ent_stringID:
                                    {
                                        pInfo.Nodes.Add(new StringIDNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_unused:
                                    {
                                        int eSize = ToInt32(xr.GetAttribute(att_size));
                                        if (tagJustRead && eSize == 8) unusedJustRead = true;

                                        pInfo.Nodes.Add(new VoidNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             eSize,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_undefined:
                                    {
                                        pInfo.Nodes.Add(new UnknownNode(
                                             man,
                                             curParent,
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             4,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_byte:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Byte,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_short:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                              man,
                                              curParent,
                                              NodeType.Short,
                                              xr.GetAttribute(att_name),
                                              ToInt32(xr.GetAttribute(att_offset)),
                                              ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_ushort:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.UShort,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_int:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Long,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_uint:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.ULong,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_float:
                                    {
                                        pInfo.Nodes.Add(new ValueNode(
                                             man,
                                             curParent,
                                             NodeType.Float,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_option:
                                    {
                                        OptionNode on = new OptionNode(
                                            xr.GetAttribute(att_name),
                                            ToInt32(xr.GetAttribute(att_value)));

                                        if (curBitmask != null) curBitmask.OptionNodes.Add(on);
                                        else if (curEnum != null) curEnum.OptionNodes.Add(on);

                                        break;
                                    }
                                case ent_bitmask8:
                                    {
                                        pInfo.Nodes.Add(curBitmask = new BitmaskNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             8,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_bitmask16:
                                    {
                                        pInfo.Nodes.Add(curBitmask = new BitmaskNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             16,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_bitmask:
                                case ent_bitmask32:
                                    {
                                        pInfo.Nodes.Add(curBitmask = new BitmaskNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             32,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_enum8:
                                    {
                                        pInfo.Nodes.Add(curEnum = new EnumNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             8,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_enum16:
                                    {
                                        pInfo.Nodes.Add(curEnum = new EnumNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             16,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_enum:
                                case ent_enum32:
                                    {
                                        pInfo.Nodes.Add(curEnum = new EnumNode(
                                             man,
                                             curParent,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             32,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_string32:
                                    {
                                        pInfo.Nodes.Add(new StringNode(
                                             man,
                                             curParent,
                                             0,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             32,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_string256:
                                    {
                                        pInfo.Nodes.Add(new StringNode(
                                             man,
                                             curParent,
                                             0,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             256,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_unicode64:
                                    {
                                        pInfo.Nodes.Add(new StringNode(
                                             man,
                                             curParent,
                                             1,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             64,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                                case ent_unicode256:
                                    {
                                        pInfo.Nodes.Add(new StringNode(
                                             man,
                                             curParent,
                                             1,
                                             xr.GetAttribute(att_name),
                                             ToInt32(xr.GetAttribute(att_offset)),
                                             256,
                                             ToBool(xr.GetAttribute(att_visible))));
                                        break;
                                    }
                            }

                            if (tagJustRead && xrNameLower != ent_tagRef)
                                tagJustRead = false;

                            if (unusedJustRead && xrNameLower != ent_unused)
                                unusedJustRead = false;

                            break;
                        }
                    case XmlNodeType.EndElement:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case ent_bitmask8:
                                case ent_bitmask16:
                                case ent_bitmask:
                                case ent_bitmask32:
                                    {
                                        curBitmask = null;
                                        break;
                                    }

                                case ent_enum8:
                                case ent_enum16:
                                case ent_enum:
                                case ent_enum32:
                                    {
                                        curEnum = null;
                                        break;
                                    }

                                case ent_struct:
                                    {
                                        curParent = curParent.Parent;
                                        break;
                                    }
                            }
                            break;
                        }
                }
            }

            xr.Close();
            return true;
        }

        private static bool ReadCSharpPlugin(PluginManager man, PluginInfo pInfo)
        {
            // Compile options and refs.
            CompilerParameters compilerParams = new CompilerParameters();
            compilerParams.GenerateExecutable = false;
            compilerParams.GenerateInMemory = false;
            compilerParams.TreatWarningsAsErrors = false;
            compilerParams.ReferencedAssemblies.Add("System.dll");
            compilerParams.ReferencedAssemblies.Add("System.Windows.Forms.dll");
            compilerParams.ReferencedAssemblies.Add(Application.StartupPath + "\\Halo3Map.dll");

            // Compile the script.
            CSharpCodeProvider codeProvider = new CSharpCodeProvider();
            CompilerResults results = codeProvider.CompileAssemblyFromFile(compilerParams, pInfo.FileName);

            // Handle any errors..
            if (results.Errors.Count != 0)
            {
                foreach (CompilerError e in results.Errors)
                {
                    if (MessageBox.Show(string.Format("\"{0}\"\r\n\r\nLine number: {1}, Error number: {2} \r\n\r\n\"{3}\"",
                        pInfo.FileName, e.Line, e.ErrorNumber, e.ErrorText), string.Empty, MessageBoxButtons.OKCancel,
                        MessageBoxIcon.Exclamation) == DialogResult.Cancel)
                        break;
                }
                return false;
            }

            // No errors, proceed to create an instance of the script class.     
            H3Tag tag = null;
            foreach (Type t in results.CompiledAssembly.GetTypes())
            {
                if (t.BaseType == typeof(H3Tag))
                {
                    tag = (H3Tag)Activator.CreateInstance(t);
                    break;
                }
            }
            if (tag == null) return false;

            Type tagType = tag.GetType();
            object[] attributes = tagType.GetCustomAttributes(typeof(TagAttribute), false);
            if (attributes.Length != 0)
            {
                TagAttribute tagAtt = (TagAttribute)attributes[0];
                pInfo.Class = tagAtt.TagType;
                pInfo.Author = tagAtt.Author;
                pInfo.Version = tagAtt.Version.ToString();
                pInfo.HeaderLength = Tags.GetMetaSize(pInfo.Class);

                pInfo.Revisions.Add(new Revision(
                    pInfo.Author, pInfo.Version, string.Empty));
            }

            // Recursive..
            ReadCSharpPlugin_Loop(man, pInfo, tagType, null);
            return true;
        }

        private static void ReadCSharpPlugin_Loop(PluginManager man, PluginInfo pInfo, Type obj, StructNode curParent)
        {
            // Only take fields which are public or instance.
            FieldInfo[] fields = obj.GetFields(BindingFlags.Public | BindingFlags.Instance);
            int curPos = 0;

            foreach (FieldInfo fi in fields)
            {
                if (fi.FieldType == typeof(byte))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             1, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Byte,
                             fi.Name, curPos, true));
                    }
                    curPos += 1;
                }
                else if (fi.FieldType == typeof(char))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             1, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new StringNode(
                             man, curParent, 0, fi.Name,
                             curPos, 1, true));
                    }
                    curPos += 1;
                }
                else if (fi.FieldType == typeof(short))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             2, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Short,
                             fi.Name, curPos, true));
                    }
                    curPos += 2;
                }
                else if (fi.FieldType == typeof(ushort))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             2, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.UShort,
                             fi.Name, curPos, true));
                    }
                    curPos += 2;
                }
                else if (fi.FieldType == typeof(int))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             4, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Long,
                             fi.Name, curPos, true));
                    }
                    curPos += 4;
                }
                else if (fi.FieldType == typeof(uint))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             4, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Long,
                             fi.Name, curPos, true));
                    }
                    curPos += 4;
                }
                else if (fi.FieldType == typeof(long))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             8, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Quad,
                             fi.Name, curPos, true));
                    }
                    curPos += 8;
                }
                else if (fi.FieldType == typeof(ulong))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             8, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.UQuad,
                             fi.Name, curPos, true));
                    }
                    curPos += 8;
                }
                else if (fi.FieldType == typeof(float))
                {
                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             4, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Float,
                             fi.Name, curPos, true));
                    }
                    curPos += 4;
                }
                else if (fi.FieldType == typeof(double))
                {

                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             8, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new ValueNode(
                             man, curParent, NodeType.Double,
                             fi.Name, curPos, true));
                    }
                    curPos += 8;
                }
                else if (fi.FieldType.IsEnum)
                {
                    Type uType = Enum.GetUnderlyingType(fi.FieldType);
                    FieldInfo[] eOptions = fi.FieldType.GetFields();
                    EnumNode eNode = null;

                    if (uType == typeof(byte))
                    {
                        eNode = new EnumNode(
                         man, curParent, fi.Name,
                         curPos, 8, true);

                        for (int x = 1; x < eOptions.Length; x++)
                        {
                            eNode.OptionNodes.Add(
                                new OptionNode(eOptions[x].Name, (byte)eOptions[x].GetRawConstantValue()));
                        }

                        curPos += 1;
                    }
                    else if (uType == typeof(short))
                    {
                        eNode = new EnumNode(
                         man, curParent, fi.Name,
                         curPos, 16, true);

                        for (int x = 1; x < eOptions.Length; x++)
                        {
                            eNode.OptionNodes.Add(
                                new OptionNode(eOptions[x].Name, (short)eOptions[x].GetRawConstantValue()));
                        }

                        curPos += 2;
                    }
                    else if (uType == typeof(int))
                    {
                        eNode = new EnumNode(
                         man, curParent, fi.Name,
                         curPos, 32, true);

                        for (int x = 1; x < eOptions.Length; x++)
                        {
                            eNode.OptionNodes.Add(
                                new OptionNode(eOptions[x].Name, (int)eOptions[x].GetRawConstantValue()));
                        }

                        curPos += 4;
                    }
                    else if (uType == typeof(long))
                    {
                        /*
                         * We won't support man for now.. not unless it's needed; too
                         * much stuff will have to be changed.. which I can't be bothered
                         * with right now..
                        
                        eNode = new EnumNode(
                         man, curParent, fi.Name,
                         curPos, 64, true);

                        for (int x = 1; x < eFields.Length; x++)
                        {
                            eNode.OptionNodes.Add(
                                new OptionNode(eFields[x].Name, (long)eFields[x].GetRawConstantValue()));
                        }
                        */

                        curPos += 8;
                    }

                    pInfo.Nodes.Add(eNode);
                }
                else if (fi.FieldType == typeof(char[]))
                {
                    int strLen = 0;

                    object[] attributes = fi.GetCustomAttributes(typeof(SizeAttribute), false);
                    if (attributes.Length != 0)
                    {
                        strLen = ((SizeAttribute)attributes[0]).Size;
                    }

                    pInfo.Nodes.Add(new StringNode(
                         man, curParent, 0, fi.Name,
                         curPos, strLen, true));

                    curPos += strLen;
                }
                else if (fi.FieldType == typeof(byte[]))
                {
                    int arrayLen = 0;

                    object[] attributes = fi.GetCustomAttributes(typeof(SizeAttribute), false);
                    if (attributes.Length != 0)
                    {
                        arrayLen = ((SizeAttribute)attributes[0]).Size;
                    }

                    if (fi.Name.IndexOf("unknown", StringComparison.OrdinalIgnoreCase) != -1)
                    {
                        pInfo.Nodes.Add(new UnknownNode(
                             man, curParent, curPos,
                             arrayLen, true));
                    }
                    else
                    {
                        pInfo.Nodes.Add(new VoidNode(
                             man, curParent, fi.Name,
                             curPos, arrayLen, true));
                    }

                    curPos += arrayLen;
                }
                else if (fi.FieldType == typeof(string))
                {
                    int strLen = 0;
                    byte encoding = 0;

                    object[] attributes = fi.GetCustomAttributes(typeof(StringAttribute), false);
                    if (attributes.Length != 0)
                    {
                        strLen = ((StringAttribute)attributes[0]).Size;
                        switch (((StringAttribute)attributes[0]).StringType)
                        {
                            case StringType.Ascii: encoding = 0; break;
                            case StringType.Unicode: encoding = 1; break;
                        }
                    }

                    pInfo.Nodes.Add(new StringNode(
                         man, curParent, encoding, fi.Name,
                         curPos, strLen, true));

                    curPos += strLen;
                }
                else if (fi.FieldType == typeof(StringID))
                {
                    pInfo.Nodes.Add(new StringIDNode(
                         man, curParent, fi.Name,
                         curPos, true));

                    curPos += 1; // ERM TODO
                }
                else if (fi.FieldType == typeof(TagReference))
                {
                    pInfo.Nodes.Add(new TagRefNode(
                         man, curParent, true, fi.Name,
                         curPos, true));

                    curPos += 16;
                }
                else if (fi.FieldType.IsGenericType && fi.FieldType.GetGenericTypeDefinition() == typeof(H3Reflexive<>))
                {
                    Type chunkType = fi.FieldType.GetGenericArguments()[0];
                    int chunkSize = -1;

                    object[] attributes = chunkType.GetCustomAttributes(typeof(SizeAttribute), false);
                    if (attributes.Length == 0)
                    {
                        MessageBox.Show(string.Format("Error: Unknown Size.\r\n\r\n\"{0}\"", chunkType.ToString()),
                            string.Empty, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    }
                    else
                    {
                        chunkSize = ((SizeAttribute)attributes[0]).Size;

                        StructNode sn = new StructNode(
                            man, curParent, fi.Name,
                            curPos, chunkSize, true);

                        pInfo.Nodes.Add(sn);
                        ReadCSharpPlugin_Loop(man, pInfo, chunkType, sn);
                    }

                    curPos += 12;
                }
            }
        }

        #endregion

        #region Display

        /// <summary>
        /// Parses the data read from the PLugin into a Control array.
        /// </summary>
        /// <param name="parentWidth">The width of the control to inherit the returned controls. </param>
        public Control[] ParseForViewing(int parentWidth)
        {
            // Get control count.
            int cCount = 0;
            for (int x = 0; x < CurrentPlugin.Nodes.Count; x++)
            {
                if (CurrentPlugin.Nodes[x].IsVisible() && CurrentPlugin.Nodes[x].Parent == null)
                    cCount += CurrentPlugin.Nodes[x].ControlCount;
            }

            // Use count to size array.
            Control[] collection = new Control[cCount];
            int cIndex = 0;

            BitReader br = new BitReader(Map.FileName, Endian.Big);
            int mOffset = Map.Tags.SelectedTag.MetaOffset;

            int xpos = 12;
            int ypos = 12;

            // Fill array.
            for (int x = 0; x < CurrentPlugin.Nodes.Count; x++)
            {
                if (CurrentPlugin.Nodes[x].IsVisible())
                {
                    // Add to base array.
                    if (CurrentPlugin.Nodes[x].Parent == null)
                    {
                        CurrentPlugin.Nodes[x].Read(br, mOffset);
                        CurrentPlugin.Nodes[x].ParseForViewing(collection, ref cIndex, xpos, ref ypos);

                        // Extra specific control re-sizes.
                        switch (CurrentPlugin.Nodes[x].NodeType)
                        {
                            case NodeType.Reflexive:
                                {
                                    collection[cIndex - 1].Width = parentWidth - xpos;
                                    break;
                                }
                            case NodeType.Bitmask:
                            case NodeType.Enum:
                                {
                                    collection[cIndex - 1].Width = parentWidth - (xpos * 2);
                                    break;
                                }
                            case NodeType.TagRef:
                                {
                                    TagRefNode node = (CurrentPlugin.Nodes[x] as TagRefNode);
                                    node.Parent_Resized(node.Parent == null ? parentWidth : node.Parent.Panel1.Width);
                                    break;
                                }
                        }
                    }
                    else if (CurrentPlugin.Nodes[x].Parent.IsVisible()) // Add to parent array.
                    {
                        if (CurrentPlugin.Nodes[x].Parent.Enabled)
                            CurrentPlugin.Nodes[x].Read(br, CurrentPlugin.Nodes[x].Parent.CurrentChunk);

                        Control[] tmpCollection = new Control[CurrentPlugin.Nodes[x].ControlCount];
                        int tmpIndx = 0;
                        int tmpY = CurrentPlugin.Nodes[x].Parent.Panel1.Height;

                        CurrentPlugin.Nodes[x].ParseForViewing(tmpCollection, ref tmpIndx, xpos, ref tmpY);
                        CurrentPlugin.Nodes[x].Parent.Panel1.Controls.AddRange(tmpCollection);

                        int diff = tmpY - CurrentPlugin.Nodes[x].Parent.Panel1.Height;
                        StructNode sn = CurrentPlugin.Nodes[x].Parent;
                        while (sn != null)
                        {
                            sn.Panel1.Height += diff;
                            sn = sn.Parent;
                        }
                        ypos += diff;

                        // Extra specific control re-sizes.
                        switch (CurrentPlugin.Nodes[x].NodeType)
                        {
                            case NodeType.Reflexive:
                                {
                                    tmpCollection[tmpIndx - 1].Width = (CurrentPlugin.Nodes[x].Parent.Panel1.Width - xpos);
                                    tmpCollection[tmpIndx - 1].Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
                                    break;
                                }
                            case NodeType.Bitmask:
                            case NodeType.Enum:
                                {
                                    tmpCollection[tmpIndx - 1].Width = (CurrentPlugin.Nodes[x].Parent.Panel1.Width - (xpos * 2));
                                    tmpCollection[tmpIndx - 1].Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
                                    break;
                                }
                            case NodeType.TagRef:
                                {
                                    TagRefNode node = (CurrentPlugin.Nodes[x] as TagRefNode);
                                    node.Parent_Resized(node.Parent == null ? parentWidth : node.Parent.Panel1.Width);
                                    break;
                                }
                        }
                    }
                }
            }

            // Close reader, return array.
            br.Close();
            return collection;
        }

        /// <summary>
        /// Refreshes the values displayed by the plugin controls.
        /// </summary>
        public void RefreshControls()
        {
            BitReader br = new BitReader(Map.FileName, Endian.Big);
            int mOffset = Map.Tags.SelectedTag.MetaOffset;

            for (int x = 0; x < CurrentPlugin.Nodes.Count; x++)
            {
                if (CurrentPlugin.Nodes[x].IsVisible())
                {
                    if (CurrentPlugin.Nodes[x].Parent == null)
                    {
                        CurrentPlugin.Nodes[x].Read(br, mOffset);
                    }
                    else if (CurrentPlugin.Nodes[x].Parent.Enabled)
                    {
                        CurrentPlugin.Nodes[x].Read(br, CurrentPlugin.Nodes[x].Parent.CurrentChunk);
                    }
                }
            }

            br.Close();
        }

        /// <summary>
        /// Resizes the plugin controls on a resize event.
        /// </summary>
        public void OnPanelResize(object sender, EventArgs e)
        {
            Panel pan = (sender as Panel);
            if (pan.Controls.Count == 0) return;

            //int newW = pan.Width - 24 - (pan.VerticalScroll.Visible ?
            //    SystemInformation.VerticalScrollBarWidth : 0);
            int newW = pan.Width - 12; // 12 = xpos

            Type typeofPanel = typeof(Panel);
            Type typeofGroupBox = typeof(GroupBox);
            Type typeofComboBox = typeof(ComboBox);

            for (int x = 0; x < pan.Controls.Count; x++)
            {
                Type typeofCurrent = pan.Controls[x].GetType();

                // Reflexive
                if (typeofCurrent == typeofPanel)
                {
                    pan.Controls[x].Width = newW;
                }
                // Enum/Bitmask
                else if (typeofCurrent == typeofGroupBox)
                {
                    pan.Controls[x].Width = newW - 12;
                }
            }

            for (int x = 0; x < CurrentPlugin.Nodes.Count; x++)
            {
                if (CurrentPlugin.Nodes[x].NodeType == NodeType.TagRef)
                {
                    TagRefNode node = (CurrentPlugin.Nodes[x] as TagRefNode);
                    node.Parent_Resized(node.Parent == null ? newW : node.Parent.Panel1.Width);
                }
            }
        }

        #endregion

        #region Write

        /// <summary>
        /// Writes the new values back to the map file.
        /// </summary>
        public void Write()
        {
            BitWriter bw = new BitWriter(Map.FileName, Endian.Big);
            int mOffset = Map.Tags.SelectedTag.MetaOffset;

            for (int x = 0; x < CurrentPlugin.Nodes.Count; x++)
            {
                if (CurrentPlugin.Nodes[x].IsVisible())
                {
                    if (CurrentPlugin.Nodes[x].Parent == null)
                    {
                        CurrentPlugin.Nodes[x].Write(bw, mOffset);
                    }
                    else if (CurrentPlugin.Nodes[x].Parent.IsVisible() &&
                        CurrentPlugin.Nodes[x].Parent.Enabled)
                    {
                        CurrentPlugin.Nodes[x].Write(bw, CurrentPlugin.Nodes[x].Parent.CurrentChunk);
                    }
                }
            }

            bw.Close();
        }

        #endregion

        #region Misc

        /// <summary>
        /// Converts a string value into a 32bit integer.
        /// </summary>
        /// <param name="value">The string to convert. </param>
        private static int ToInt32(string value)
        {
            try { return int.Parse(value); }
            catch { return 0; }
        }

        /// <summary>
        /// Converts a string value into a boolean.
        /// </summary>
        /// <param name="value">The string to convert. </param>
        private static bool ToBool(string value)
        {
            try { return bool.Parse(value); }
            catch { return true; }
        }

        /// <summary>
        /// Finds out whether the specified type is visible on the Meta Editor.
        /// </summary>
        /// <param name="type">The specified type. </param>
        public static bool IsVisible(NodeType type)
        {
            return ((type & VisibleMask) == type);
        }

        /// <summary>
        /// Gets the point at which nodes using textboxes or comboboxes should
        /// place them on the x-axis.
        /// </summary>
        private static void GetBoxUsersStart()
        {
            BoxUsersStart = 67;
            return;

            // Get strings.
            string[] enumNames = Enum.GetNames(typeof(NodeType));
            // Index of Longest.
            int index = -1;
            // Loop.
            for (int x = 1; x < enumNames.Length; x++)
            {
                // Convert back to NodeType.
                NodeType nt = (NodeType)Enum.Parse(typeof(NodeType), enumNames[x]);
                // Filter.
                if (nt != NodeType.None && nt != NodeType.BoxUsers && (NodeType.BoxUsers & nt) == nt)
                {
                    // Hacky.. but w/e..
                    if (nt == NodeType.TagRef) enumNames[x] = "Tag Ref";

                    if (index == -1 || enumNames[x].Length > enumNames[index].Length)
                        index = x;
                }
            }
            if (index == -1) return;

            Label label = new Label();
            label.Text = enumNames[index];
            BoxUsersStart = label.PreferredWidth + 2;
            label.Dispose();
        }

        #endregion

        #region Convert

        /// <summary>
        /// Converts one plugin type to another.
        /// </summary>
        /// <param name="inputFileName">The input filename. </param>
        /// <param name="inputType">The input type. </param>
        /// <param name="outputFileName">The output filename. </param>
        /// <param name="outputType">The output type. </param>
        public static void ConvertPlugin(string inputFileName, PluginType inputType,
            string outputFileName, PluginType outputType)
        {
            PluginConverter.ConvertInfo.Clear();
            PluginConverter.ConvertInfo.FileName = inputFileName;

            // Load.
            switch (inputType)
            {
                case PluginType.DTA:
                    ReadDefinedThroughAttributesPlugin(null, PluginConverter.ConvertInfo);
                    break;

                case PluginType.ENT:
                    ReadEntityPlugin(null, PluginConverter.ConvertInfo);
                    break;

                case PluginType.CS:
                    ReadCSharpPlugin(null, PluginConverter.ConvertInfo);
                    break;
            }

            // Save.
            switch (outputType)
            {
                case PluginType.DTA:
                    PluginConverter.SaveAsDefinedThroughAttributesPlugin(outputFileName);
                    break;

                case PluginType.ENT:
                    PluginConverter.SaveAsEntityPlugin(outputFileName);
                    break;

                case PluginType.CS:
                    PluginConverter.SaveAsCSharpPlugin(outputFileName);
                    break;
            }
        }

        #endregion

        /// <summary>
        /// The plugin types.
        /// </summary>
        public enum PluginType
        {
            /// <summary>
            /// Defined Through Attributes
            /// </summary>
            DTA,
            /// <summary>
            /// Entity
            /// </summary>
            ENT,
            /// <summary>
            /// CSharp
            /// </summary>
            CS
        }

        /// <summary>
        /// Stores info on a plugin.
        /// </summary>
        public class PluginInfo
        {
            #region Instance Data

            private string __fileName__;
            private string __class__;
            private string __author__;
            private string __version__;
            private int __headerLength__;
            private List<Revision> __revisions__;
            private List<BaseNode> __nodes__;

            #endregion

            #region Accessors

            /// <summary>
            /// FileName
            /// </summary>
            public string FileName { get { return __fileName__; } set { __fileName__ = value; } }
            /// <summary>
            /// Class
            /// </summary>
            public string Class { get { return __class__; } set { __class__ = value; } }
            /// <summary>
            /// Author
            /// </summary>
            public string Author { get { return __author__; } set { __author__ = value; } }
            /// <summary>
            /// Version
            /// </summary>
            public string Version { get { return __version__; } set { __version__ = value; } }
            /// <summary>
            /// HeaderLength
            /// </summary>
            public int HeaderLength { get { return __headerLength__; } set { __headerLength__ = value; } }
            /// <summary>
            /// Revisions
            /// </summary>
            public List<Revision> Revisions { get { return __revisions__; } set { __revisions__ = value; } }
            /// <summary>
            /// Nodes
            /// </summary>
            public List<BaseNode> Nodes { get { return __nodes__; } set { __nodes__ = value; } }

            #endregion

            public PluginInfo()
            {
                Revisions = new List<Revision>();
                Nodes = new List<BaseNode>();
            }

            public void Clear()
            {
                Class = string.Empty;
                Author = string.Empty;
                Version = string.Empty;
                HeaderLength = 0;
                Revisions.Clear();
                Nodes.Clear();
            }
        }
    }
}