using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.IO;
using System.Extended;
using H3UtilityToolkit;
using H3UtilityToolkit.Essentials;
using H3UtilityToolkit.Misc;

namespace H3UtilityToolkit.Structures
{
    /// <summary>
    /// Scenario
    /// </summary>
    public class Scnr
    {
        #region Instance Data

        private Map __map__;
        private int __identity__;

        public int tSize;
        public int tOffset;
        public Script[] scripts;
        public ScriptSyntax[] scriptSyntaxes;

        #endregion

        #region Accessors

        /// <summary>
        /// Map reference, used when in need of other functions outside of this class.
        /// </summary>
        public Map Map { get { return __map__; } set { __map__ = value; } }
        /// <summary>
        /// The identity of the tag to be used by this class.
        /// </summary>
        public int Identity { get { return __identity__; } set { __identity__ = value; } }

        #endregion

        /// <summary>
        /// Initializes a new instance of the H3UtilityToolkit.Structures.Scnr class based 
        /// on the supplied map reference and tag identity.
        /// </summary>
        /// <param name="map">The map reference. </param>
        /// <param name="identity">The identity of the tag to be used by this class. </param>
        public Scnr(Map map, int identity)
        {
            this.Map = map;
            this.Identity = identity;
        }

        #region [scnr] Related..

        public void LoadScriptTable()
        {
            BitReader br = new BitReader(Map.FileName, Endian.Big);
            LoadScriptTable(br);
            br.Close();
        }

        public void LoadScriptTable(BitReader br)
        {
            int tagIndx = Map.Tags.IndexOf(Identity);
            if (tagIndx == -1) return;

            br.BaseStream.Position = Map.Tags.TagsList[tagIndx].MetaOffset
                + 992;

            // Table
            tSize = br.ReadInt32();
            br.BaseStream.Position += 8;
            tOffset = br.ReadInt32() - Map.Header.MapMagic;

            // Scripts
            br.BaseStream.Position += 4;
            int cc = br.ReadInt32();
            int tr = br.ReadInt32() - Map.Header.MapMagic;

            br.BaseStream.Position = tr;
            scripts = new Script[cc];
            for (int x = 0; x < cc; x++)
            {
                Script script = new Script();
                script.Name = br.ReadCString(32);
                script.ScriptType = (ScriptType)br.ReadInt16();
                script.ReturnType = (ReturnType)br.ReadInt16();
                script.ExpressionIndex = br.ReadInt16();
                script.ChunkIndex = br.ReadInt16();
                br.BaseStream.Position += 12;
                scripts[x] = script;
            }

            // Script Syntax
            br.BaseStream.Position = Map.Tags.TagsList[tagIndx].MetaOffset
                + 1188;
            cc = br.ReadInt32();
            tr = br.ReadInt32() - Map.Header.MapMagic;

            br.BaseStream.Position = tr;
            scriptSyntaxes = new ScriptSyntax[cc];
            for (int x = 0; x < cc; x++)
            {
                ScriptSyntax ss = new ScriptSyntax();
                ss.ExpressionID = br.ReadInt16();
                ss.Identity = br.ReadInt16();
                ss.ValueType = br.ReadInt16();
                ss.ExpressionType = br.ReadInt16();
                ss.SiblingPointer = br.ReadInt16();
                ss.SiblingIndex = br.ReadInt16();
                ss.ScriptStringOffset = br.ReadInt32();
                ss.Value = br.ReadBytes(4);
                br.BaseStream.Position += 4;
                scriptSyntaxes[x] = ss;
            }
        }

        public ListViewItem[] ParseForListview()
        {
            // Gather the strings
            ListViewItem[] items = new ListViewItem[scripts.Length];
            for (int x = 0; x < scripts.Length; x++)
            {
                items[x] = new ListViewItem(scripts[x].Name);
                items[x].SubItems.Add(scripts[x].ScriptType.ToString());
                items[x].SubItems.Add(scripts[x].ReturnType.ToString());
            }
            return items;
        }

        public struct Script
        {
            public string Name;
            public ScriptType ScriptType;
            public ReturnType ReturnType;
            public short ExpressionIndex;
            public short ChunkIndex;
        }

        public struct ScriptSyntax
        {
            public short ExpressionID;
            public short Identity;
            public short ValueType;
            public short ExpressionType;
            public short SiblingPointer;
            public short SiblingIndex;
            public int ScriptStringOffset;
            public byte[] Value;
        }

        public enum ScriptType : short
        {
            Startup,
            Dormant,
            Continuous,
            Static,
            Stub,
            Command_Script
        }

        public enum ReturnType : short
        {
            UnParsed,
            SpecialForm,
            FunctionName,
            PassThrough,
            Void,
            Boolean,
            Real,
            Short,
            Long,
            String,
            script,
            string_id,
            unit_seat_mapping,
            trigger_volume,
            cutscene_flag,
            cutscene_camera_point,
            cutscene_title,
            cutscene_recording,
            device_group,
            ai,
            ai_command_list,
            ai_command_script,
            ai_behaviour,
            ai_orders,
            starting_profile,
            conversation,
            structure_bsp,
            navpoint,
            point_reference,
            style,
            hud_message,
            object_list,
            scenary,
            effect,
            unit,
            looping_sound,
            animation_graph,
            object_definition,
            bitmap,
            shader,
            UNK,
            render_model,
            structure_definition,
            lightmap_definition,
            game_difficulty,
            actor_type,
            hud_corner,
            model_state,
            value,
            network_event,
            value2,
            unit_name,
            vehicle_name,
            weapon_name,
            device_name,
            scenery_name,
            object_name
        }

        #endregion
    }
}