using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.IO;
using System.Extended;
using H3UtilityToolkit;
using H3UtilityToolkit.Essentials;
using H3UtilityToolkit.Misc;

namespace H3UtilityToolkit.Structures
{
    /// <summary>
    /// Resources
    /// </summary>
    public class Zone
    {
        #region Instance Data

        private Map __map__;
        private int __identity__;

        #endregion

        #region Accessors

        /// <summary>
        /// Map reference, used when in need of other functions outside of this class.
        /// </summary>
        public Map Map { get { return __map__; } set { __map__ = value; } }
        /// <summary>
        /// The identity of the tag to be used by this class.
        /// </summary>
        public int Identity { get { return __identity__; } set { __identity__ = value; } }

        #endregion

        /// <summary>
        /// Initializes a new instance of the H3UtilityToolkit.Structures.Zone class based 
        /// on the supplied map reference and tag identity.
        /// </summary>
        /// <param name="map">The map reference. </param>
        /// <param name="identity">The identity of the tag to be used by this class. </param>
        public Zone(Map map, int identity)
        {
            this.Map = map;
            this.Identity = identity;
        }

        #region [zone] Related..

        public ListViewItem[] GetAssetTable()
        {
            BitReader br = new BitReader(Map.FileName, Endian.Big);
            ListViewItem[] items = GetAssetTable(br);
            br.Close();
            return items;
        }

        public ListViewItem[] GetAssetTable(BitReader br)
        {
            int tagIndx = Map.Tags.IndexOf(Identity);
            if (tagIndx == -1) return null;

            br.BaseStream.Position = Map.Tags.TagsList[tagIndx].MetaOffset + 88;

            int cc = br.ReadInt32();
            int tr = br.ReadInt32() - Map.Header.MapMagic;

            List<AssetBlock> abs = new List<AssetBlock>(cc);
            br.BaseStream.Position = tr;

            for (int x = 0; x < cc; x++)
            {
                AssetBlock ab = new AssetBlock();

                string tt = br.ReadString(4);
                br.BaseStream.Position += 8;
                int ident = br.ReadInt32();

                if (ident == -1)
                {
                    ab.Index = x;
                    ab.Type = Tags.NullTagType;
                    ab.Path = Tags.NullTagPath;
                    ab.Pointer1 = -1;
                    ab.Pointer2 = -1;
                    br.BaseStream.Position += 48;
                }
                else
                {
                    ab.Index = x;

                    ab.Type = tt;
                    ab.Path = Map.Tags.TagsList[Map.Tags.IndexOf(ident)].Path;

                    br.BaseStream.Position += 16;
                    short rawCountFlag = br.ReadInt16();

                    br.BaseStream.Position += 10;
                    ab.Pointer1 = br.ReadInt32() - Map.Header.MapMagic;

                    br.BaseStream.Position += 8;
                    ab.Pointer2 = br.ReadInt32() - Map.Header.MapMagic;
                    if (rawCountFlag == 1) ab.Pointer2 = -1;

                    br.BaseStream.Position += 4;
                }

                abs.Add(ab);
            }

            // Sort
            List<AssetBlock> tmpList = new List<AssetBlock>(cc);
            while (abs.Count != 0)
            {
                int sIndx = 0;
                for (int x = 1; x < abs.Count; x++)
                {
                    if (abs[x].Pointer1 < abs[sIndx].Pointer1) sIndx = x;
                }
                tmpList.Add(abs[sIndx]);
                abs.RemoveAt(sIndx);
            }
            abs = tmpList;

            ListViewItem[] items = new ListViewItem[cc];
            for (int x = 0; x < cc; x++)
            {
                items[x] = new ListViewItem(abs[x].Index.ToString());
                items[x].SubItems.Add(abs[x].Type);
                items[x].SubItems.Add(abs[x].Path);
                items[x].SubItems.Add(abs[x].Pointer1.ToString());
                items[x].SubItems.Add(abs[x].Pointer2.ToString());
            }
            return items;
        }

        public struct AssetBlock
        {
            public int Index;
            public string Type;
            public string Path;
            public int Pointer1; 
            public int Pointer2;
        }

        #endregion
    }
}