//--------------------------------------------------------------------------------------
// File: BitReader.cs
//
// Reads primitive data types as binary values in a specific endian and/or encoding
//
// Copyright (c) Alex Lynch (alias. Prey). All rights reserved.
//--------------------------------------------------------------------------------------
using System.Text;
using System.IO;

namespace System.Extended
{
    ///<summary>
    /// The data's byte order.
    ///</summary>
    public enum Endian
    {
        ///<summary>
        /// [Intel] Front-to-Back.
        ///</summary>
        Little,
        ///<summary>
        /// [Motorola] Back-to-Front.
        ///</summary>
        Big
    }

    ///<summary>
    /// Reads primitive data types as binary values in a specific endian and/or encoding.
    ///</summary>
    public class BitReader : BinaryReader
    {
        #region Instance Data

        private Endian __endian__;

        #endregion

        #region Accessors

        ///<summary>
        /// The stream's byte order.
        ///</summary>
        public Endian Endian { get { return __endian__; } set { __endian__ = value; } }

        #endregion

        /// <summary>
        /// Initialises a new instance of the System.Extended.BitReader class based on the
        /// supplied file name and specific file endian.
        /// </summary>
        /// <param name="fileName">The supplied file name. </param>
        /// <param name="endian">The specific file endian. </param>
        public BitReader(string fileName, Endian endian)
            : base(new FileStream(fileName, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite))
        {
            this.Endian = endian;
        }

        /// <summary>
        /// Initialises a new instance of the System.Extended.BitReader class based on the
        /// supplied stream and specific file endian.
        /// </summary>
        /// <param name="stream">The supplied stream. </param>
        /// <param name="endian">The specific file endian. </param>
        public BitReader(Stream stream, Endian endian)
            : base(stream)
        {
            this.Endian = endian;
        }

        /// <summary>
        /// Reads the next character from the current stream and advances the current
        /// position of the stream by one byte.
        /// </summary>
        /// <returns>A character read from the current stream. </returns>
        public override char ReadChar()
        {
            return Convert.ToChar(this.ReadByte());
        }

        /// <summary>
        /// Reads count characters from the current stream, returns the data in a character
        /// array, and advances the current position by count bytes.
        /// </summary>
        /// <param name="count">The number of characters to read. </param>
        /// <returns>
        /// A character array containing data read from the underlying stream. This might
        /// be less than the number of characters requested if the end of the stream
        /// is reached.
        /// </returns>
        public override char[] ReadChars(int count)
        {
            int remaining = (int)(this.BaseStream.Length - this.BaseStream.Position);
            if (remaining < count)
            {
                return Encoding.ASCII.GetChars(this.ReadBytes(remaining));
            }
            else
            {
                return Encoding.ASCII.GetChars(this.ReadBytes(count));
            }
        }

        /// <summary>
        /// Reads count characters from the current stream, returns the data in a reversed
        /// character array, and advances the current position by count bytes.
        /// </summary>
        /// <param name="count">The number of characters to read. </param>
        /// <returns>
        /// A reversed character array containing data read from the underlying stream. This 
        /// might be less than the number of characters requested if the end of the stream
        /// is reached.
        /// </returns>
        public char[] ReadBTFChars(int count)
        {
            int remaining = (int)(this.BaseStream.Length - this.BaseStream.Position);
            char[] chars;
            if (remaining < count)
            {
                chars = Encoding.ASCII.GetChars(this.ReadBytes(remaining));
            }
            else
            {
                chars = Encoding.ASCII.GetChars(this.ReadBytes(count));
            }
            Array.Reverse(chars);
            return chars;
        }

        /// <summary>
        /// Reads a string from the current stream, and advances the current position by 
        /// the length of the string in bytes.
        /// </summary>
        /// <returns>A string read from the current stream. </returns>
        public override string ReadString()
        {
            StringBuilder sb = new StringBuilder();
            char ch;
            long pos = this.BaseStream.Position;
            long len = this.BaseStream.Length;
            while (pos != len && (ch = Convert.ToChar(this.ReadByte())) >= char.MinValue && ch <= char.MaxValue)
            {
                sb.Append(ch);
                pos++;
            }
            return sb.ToString();
        }

        /// <summary>
        /// Reads a reversed string from the current stream, and advances the current 
        /// position by the length of the string in bytes.
        /// </summary>
        /// <returns>A reversed string read from the current stream. </returns>
        public string ReadBTFString()
        {
            char ch;
            long oldpos = this.BaseStream.Position;
            long pos = oldpos;
            long len = this.BaseStream.Length;
            int readlen = 0;
            while (pos != len && (ch = Convert.ToChar(this.ReadByte())) >= char.MinValue && ch <= char.MaxValue)
            {
                readlen++;
                pos++;
            }
            if (readlen == 0 && pos == len) readlen = (int)(len - pos);
            this.BaseStream.Position = oldpos;
            char[] chars = Encoding.ASCII.GetChars(this.ReadBytes(readlen));
            Array.Reverse(chars);
            return new string(chars);
        }

        /// <summary>
        /// Reads a string of the specified length from the current stream, using
        /// System.Text.Encoding.ASCII, and advances the current position by the 
        /// length in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <returns>
        /// A string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadString(int length)
        {
            return ReadString(length, Encoding.ASCII);
        }

        /// <summary>
        /// Reads a reversed string of the specified length from the current 
        /// stream, using System.Text.Encoding.ASCII, and advances the current 
        /// position by the length in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <returns>
        /// A reversed string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadBTFString(int length)
        {
            return ReadBTFString(length, Encoding.ASCII);
        }

        /// <summary>
        /// Reads a string of the specified length from the current stream, using
        /// the specific encoding, and advances the current position by the 
        /// length in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <param name="encoding">The specific encoding. </param>
        /// <returns>
        /// A string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadString(int length, Encoding encoding)
        {
            if (length <= 0) return string.Empty;

            int remaining = (int)(this.BaseStream.Length - this.BaseStream.Position);
            if (remaining < length)
            {
                return encoding.GetString(this.ReadBytes(remaining));
            }
            else
            {
                return encoding.GetString(this.ReadBytes(length));
            }
        }

        /// <summary>
        /// Reads a reversed string of the specified length from the current 
        /// stream, using the specific encoding, and advances the current 
        /// position by the length in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <param name="encoding">The specific encoding. </param>
        /// <returns>
        /// A reversed string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadBTFString(int length, Encoding encoding)
        {
            if (length <= 0) return string.Empty;

            int remaining = (int)(this.BaseStream.Length - this.BaseStream.Position);
            char[] chars;
            if (remaining < length)
            {
                chars = encoding.GetChars(this.ReadBytes(remaining));
            }
            else
            {
                chars = encoding.GetChars(this.ReadBytes(length));
            }
            Array.Reverse(chars);
            return new string(chars);
        }

        /// <summary>
        /// Reads a string until a null terminator is read from the current stream, and 
        /// advances the current position by the length of the string in bytes.
        /// </summary>
        /// <returns>A null-terminated string read from the current stream. </returns>
        public string ReadCString()
        {
            return ReadCString('\0');
        }

        /// <summary>
        /// Reads a reversed string until a null terminator is read from the current stream, and 
        /// advances the current position by the length of the string in bytes.
        /// </summary>
        /// <returns>A null-terminated reversed string read from the current stream. </returns>
        public string ReadBTFCString()
        {
            return ReadBTFCString('\0');
        }

        /// <summary>
        /// Reads a string until the specified terminator is read from the current stream, and 
        /// advances the current position by the length of the string in bytes.
        /// </summary>
        /// <param name="nullTerminator">The specified terminator. </param>
        /// <returns>A terminated string read from the current stream. </returns>
        public string ReadCString(char terminator)
        {
            StringBuilder sb = new StringBuilder();
            char ch;
            long pos = this.BaseStream.Position;
            long len = this.BaseStream.Length;
            while (pos != len && (ch = Convert.ToChar(this.ReadByte())) != terminator)
            {
                sb.Append(ch);
            }
            return sb.ToString();
        }

        /// <summary>
        /// Reads a reversed string until the specified terminator is read from the current 
        /// stream, and advances the current position by the length of the string in bytes.
        /// </summary>
        /// <param name="nullTerminator">The specified terminator. </param>
        /// <returns>A terminated reversed string read from the current stream. </returns>
        public string ReadBTFCString(char terminator)
        {
            long pos = this.BaseStream.Position;
            long startPos = pos;
            long len = this.BaseStream.Length;
            int stringLen = 0;
            char ch;
            while (pos != len && (ch = Convert.ToChar(this.ReadByte())) != terminator)
            {
                stringLen++;
                pos++;
            }
            if (stringLen == 0) return string.Empty;
            else
            {
                this.BaseStream.Position = startPos;
                char[] chars = Encoding.ASCII.GetChars(this.ReadBytes(stringLen));
                this.BaseStream.Position++; // Term.
                Array.Reverse(chars);
                return new string(chars);
            }
        }

        /// <summary>
        /// Reads a string of the specified length until a null terminator is read from the 
        /// current stream, using System.Text.Encoding.ASCII, and advances the current 
        /// position by the length of the string in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <returns>
        /// A null-terminated string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadCString(int length)
        {
            return ReadCString(length, Encoding.ASCII, '\0');
        }

        /// <summary>
        /// Reads a reversed string of the specified length until a null terminator is read from the 
        /// current stream, using System.Text.Encoding.ASCII, and advances the current position by 
        /// the length of the string in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <returns>
        /// A null-terminated reversed string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadBTFCString(int length)
        {
            return ReadBTFCString(length, Encoding.ASCII, '\0');
        }

        /// <summary>
        /// Reads a string of the specified length until a null terminator is read from the 
        /// current stream, using the specific encoding, and advances the current 
        /// position by the length of the string in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <param name="encoding">The specific encoding. </param>
        /// <returns>
        /// A null-terminated string read from the underlying stream. This might be less than 
        /// the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadCString(int length, Encoding encoding)
        {
            return ReadCString(length, encoding, '\0');
        }

        /// <summary>
        /// Reads a reversed string of the specified length until a null terminator is read 
        /// from the current stream, using the specific encoding, and advances the current 
        /// position by the length of the string in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <param name="encoding">The specific encoding. </param>
        /// <returns>
        /// A null-terminated reversed string read from the underlying stream. This might be 
        /// less than the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadBTFCString(int length, Encoding encoding)
        {
            return ReadBTFCString(length, encoding, '\0');
        }

        /// <summary>
        /// Reads a string of the specified length until the specified terminator is read 
        /// from the current stream, using the specific encoding, and advances the current 
        /// position by the length of the string in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <param name="encoding">The specific encoding. </param>
        /// <param name="terminator">The specified terminator. </param>
        /// <returns>
        /// A terminated string read from the underlying stream. This might be 
        /// less than the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadCString(int length, Encoding encoding, char terminator)
        {
            if (length <= 0) return string.Empty;

            int remaining = (int)(this.BaseStream.Length - this.BaseStream.Position);
            if (remaining < length) length = remaining;

            byte[] buffer = this.ReadBytes(length);
            for (int x = 0; x < length; x++)
            {
                if (buffer[x] == terminator)
                {
                    return encoding.GetString(buffer, 0, x);
                }
            }
            return encoding.GetString(buffer);
        }

        /// <summary>
        /// Reads a string of the specified length until the specified terminator is read 
        /// from the current stream, using the specific encoding, and advances the current 
        /// position by the length of the string in bytes.
        /// </summary>
        /// <param name="length">The specified length. </param>
        /// <param name="encoding">The specific encoding. </param>
        /// <param name="terminator">The specified terminator. </param>
        /// <returns>
        /// A terminated string read from the underlying stream. This might be 
        /// less than the number of characters requested if the end of the stream is reached.
        /// </returns>
        public string ReadBTFCString(int length, Encoding encoding, char terminator)
        {
            if (length <= 0) return string.Empty;

            int remaining = (int)(this.BaseStream.Length - this.BaseStream.Position);
            if (remaining < length) length = remaining;

            byte[] buffer = this.ReadBytes(length);
            Array.Reverse(buffer);
            for (int x = length - 1; x >= 0; x--)
            {
                if (buffer[x] == terminator)
                {
                    return encoding.GetString(buffer, x + 1, length - x - 1);
                }
            }
            return encoding.GetString(buffer);
        }

        /// <summary>
        /// Reads a 2-byte signed integer from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by two bytes.
        /// </summary>
        /// <returns>A 2-byte signed integer read from the current stream. </returns>
        public override short ReadInt16()
        {
            if (Endian == Endian.Little) return base.ReadInt16();
            else
            {
                byte[] buffer = this.ReadBytes(2);
                Array.Reverse(buffer);
                return BitConverter.ToInt16(buffer, 0);
            }
        }

        /// <summary>
        /// Reads a 4-byte signed integer from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by four bytes.
        /// </summary>
        /// <returns>A 4-byte signed integer read from the current stream. </returns>
        public override int ReadInt32()
        {
            if (Endian == Endian.Little) return base.ReadInt32();
            else
            {
                byte[] buffer = this.ReadBytes(4);
                Array.Reverse(buffer);
                return BitConverter.ToInt32(buffer, 0);
            }
        }

        /// <summary>
        /// Reads a 8-byte signed integer from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by eight bytes.
        /// </summary>
        /// <returns>A 8-byte signed integer read from the current stream. </returns>
        public override long ReadInt64()
        {
            if (Endian == Endian.Little) return base.ReadInt64();
            else
            {
                byte[] buffer = this.ReadBytes(8);
                Array.Reverse(buffer);
                return BitConverter.ToInt64(buffer, 0);
            }
        }

        /// <summary>
        /// Reads a 2-byte unsigned integer from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by two bytes.
        /// </summary>
        /// <returns>A 2-byte unsigned integer read from the current stream. </returns>
        public override ushort ReadUInt16()
        {
            if (Endian == Endian.Little) return base.ReadUInt16();
            else
            {
                byte[] buffer = this.ReadBytes(2);
                Array.Reverse(buffer);
                return BitConverter.ToUInt16(buffer, 0);
            }
        }

        /// <summary>
        /// Reads a 4-byte unsigned integer from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by four bytes.
        /// </summary>
        /// <returns>A 4-byte unsigned integer read from the current stream. </returns>
        public override uint ReadUInt32()
        {
            if (Endian == Endian.Little) return base.ReadUInt32();
            else
            {
                byte[] buffer = this.ReadBytes(4);
                Array.Reverse(buffer);
                return BitConverter.ToUInt32(buffer, 0);
            }
        }

        /// <summary>
        /// Reads a 8-byte unsigned integer from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by eight bytes.
        /// </summary>
        /// <returns>A 8-byte unsigned integer read from the current stream. </returns>
        public override ulong ReadUInt64()
        {
            if (Endian == Endian.Little) return base.ReadUInt64();
            else
            {
                byte[] buffer = this.ReadBytes(8);
                Array.Reverse(buffer);
                return BitConverter.ToUInt64(buffer, 0);
            }
        }

        /// <summary>
        /// Reads a 4-byte floating point value from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by four bytes.
        /// </summary>
        /// <returns>A 4-byte floating point value read from the current stream. </returns>
        public override float ReadSingle()
        {
            if (Endian == Endian.Little) return base.ReadSingle();
            else
            {
                byte[] buffer = this.ReadBytes(4);
                Array.Reverse(buffer);
                return BitConverter.ToSingle(buffer, 0);
            }
        }

        ///<summary>
        /// Reads an 8-byte floating point value from the current stream in accordance with the 
        /// Endian used and advances the current position of the stream by eight bytes.
        ///</summary>
        /// <returns>An 8-byte floating point value read from the current stream. </returns>
        public override double ReadDouble()
        {
            if (Endian == Endian.Little) return base.ReadDouble();
            else
            {
                byte[] buffer = this.ReadBytes(8);
                Array.Reverse(buffer);
                return BitConverter.ToDouble(buffer, 0);
            }
        }
    }
}