//--------------------------------------------------------------------------------------
// File: BitWriter.cs
//
// Writes primitive types in binary to a stream in a specific endian and supports 
// writing strings
//
// Copyright (c) Alex Lynch (alias. Prey). All rights reserved.
//--------------------------------------------------------------------------------------
using System.IO;

namespace System.Extended
{
    ///<summary>
    /// Writes primitive types in binary to a stream in a specific endian and supports 
    /// writing strings.
    ///</summary>
    public class BitWriter : BinaryWriter
    {
        #region Instance Data

        private Endian __endian__;

        #endregion

        #region Accessors

        ///<summary>
        /// The stream's byte order.
        ///</summary>
        public Endian Endian { get { return __endian__; } set { __endian__ = value; } }

        #endregion

        /// <summary>
        /// Initialises a new instance of the System.Extended.BitWriter class based on the
        /// supplied file name and specific file endian.
        /// </summary>
        /// <param name="fileName">The supplied file name. </param>
        /// <param name="endian">The specific file endian. </param>
        public BitWriter(string fileName, Endian endian)
            : base(new FileStream(fileName, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite))
        {
            this.Endian = endian;
        }

        /// <summary>
        /// Initialises a new instance of the System.Extended.BitWriter class based on the
        /// supplied stream and specific file endian.
        /// </summary>
        /// <param name="stream">The supplied stream. </param>
        /// <param name="endian">The specific file endian. </param>
        public BitWriter(Stream stream, Endian endian)
            : base(stream)
        {
            this.Endian = endian;
        }

        /// <summary>
        /// Writes a string to this stream, and advances the current position of the stream
        /// by the length of the string in bytes.
        /// </summary>
        /// <param name="value">The value to write. </param>
        public override void Write(string value)
        {
            this.Write(value.ToCharArray());
        }

        /// <summary>
        /// Reverses and writes a string to this stream, and advances the current position of the stream
        /// by the length of the string in bytes.
        /// </summary>
        /// <param name="value">The value to write. </param>
        public void WriteBTFString(string value)
        {
            char[] chars = value.ToCharArray();
            Array.Reverse(chars);
            this.Write(chars);
        }

        /// <summary>
        /// Writes a string to this stream, shortening or padding it by the specified length, and advances 
        /// the current position of the stream by the length of the string in bytes.
        /// </summary>
        /// <param name="value">The value to write. </param>
        /// <param name="length">The specified length. </param>
        public void Write(string value, int length)
        {
            if (value.Length > length)
            {
                this.Write(value.Substring(length).ToCharArray());
            }
            else if (value.Length < length)
            {
                this.Write(value.ToCharArray());
                this.Write(new byte[length - value.Length]);
            }
            else
            {
                this.Write(value.ToCharArray());
            }
        }

        /// <summary>
        /// Reverses and writes a string to this stream, shortening or padding it by the specified length, 
        /// and advances the current position of the stream by the length of the string in bytes.
        /// </summary>
        /// <param name="value">The value to write. </param>
        /// <param name="length">The specified length. </param>
        public void WriteBTFString(string value, int length)
        {
            char[] chars = value.ToCharArray();
            Array.Reverse(chars);

            if (value.Length > length)
            {
                this.Write(chars, 0, length);
            }
            else if (value.Length < length)
            {
                this.Write(chars);
                this.Write(new byte[length - value.Length]);
            }
            else
            {
                this.Write(chars);
            }
        }

        /// <summary>
        /// Writes a byte array, shortening or padding it by the specified length, 
        /// to the underlying stream.
        /// </summary>
        /// <param name="buffer">A byte array containing the data to write. </param>
        /// <param name="length">The specified length. </param>
        public void Write(byte[] buffer, int length)
        {
            if (buffer.Length > length)
            {
                this.Write(buffer, 0, length);
            }
            else if (buffer.Length < length)
            {
                this.Write(buffer);
                this.Write(new byte[length - buffer.Length]);
            }
            else
            {
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a two-byte signed integer to the current stream in accordance with the 
        /// Endian used and advances the stream position by two bytes.
        /// </summary>
        /// <param name="value">The two-byte signed integer to write. </param>
        public override void Write(short value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a four-byte signed integer to the current stream in accordance with the 
        /// Endian used and advances the stream position by four bytes.
        /// </summary>
        /// <param name="value">The four-byte signed integer to write. </param>
        public override void Write(int value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a eight-byte signed integer to the current stream in accordance with the 
        /// Endian used and advances the stream position by eight bytes.
        /// </summary>
        /// <param name="value">The eight-byte signed integer to write. </param>
        public override void Write(long value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a two-byte unsigned integer to the current stream in accordance with the 
        /// Endian used and advances the stream position by two bytes.
        /// </summary>
        /// <param name="value">The two-byte unsigned integer to write. </param>
        public override void Write(ushort value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a four-byte unsigned integer to the current stream in accordance with the 
        /// Endian used and advances the stream position by four bytes.
        /// </summary>
        /// <param name="value">The four-byte unsigned integer to write. </param>
        public override void Write(uint value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a eight-byte unsigned integer to the current stream in accordance with the 
        /// Endian used and advances the stream position by eight bytes.
        /// </summary>
        /// <param name="value">The eight-byte unsigned integer to write. </param>
        public override void Write(ulong value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a four-byte floating-point value to the current stream in accordance with the 
        /// Endian used and advances the stream position by four bytes.
        /// </summary>
        /// <param name="value">The four-byte floating-point value to write. </param>
        public override void Write(float value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }

        /// <summary>
        /// Writes a eight-byte floating-point value to the current stream in accordance with the 
        /// Endian used and advances the stream position by eight bytes.
        /// </summary>
        /// <param name="value">The eight-byte floating-point value to write. </param>
        public override void Write(double value)
        {
            if (Endian == Endian.Little) base.Write(value);
            else
            {
                byte[] buffer = BitConverter.GetBytes(value);
                Array.Reverse(buffer);
                this.Write(buffer);
            }
        }
    }
}